"""
"""
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QLabel, \
    QTextBrowser, QScrollBar, QPushButton

from . dialogs import SettingsDialog
from .misc import heading, LabeledNumberInput, LabeledNumberInputTwoValues, \
    LabeledInput


class MainWidget(QWidget):

    def __init__(self, parent, settings, number, nsn_factory):
        assert isinstance(settings, SettingsDialog)
        super(MainWidget, self).__init__(parent)
        self.settings = settings

        self.initUI(number, nsn_factory)

    def initUI(self, number, nsn_factory):

        sdr = nsn_factory.distr_diameter
        self.btn_generate = QPushButton("Generate new array")
        self.number = LabeledNumberInput("Number", number)
        self.number2 = LabeledNumberInput("Number 2", 5)
        self.target_area_radius = LabeledNumberInput("Max radius", nsn_factory.target_area_radius)
        self.item_diameter_mean = LabeledNumberInput("Mean diameter", sdr.mu)
        self.item_diameter_std = LabeledNumberInput("Diameter range std", sdr.sigma)
        self.item_diameter_range = LabeledNumberInputTwoValues("Diameter range from",
                                                               value1=sdr.min_max[0],
                                                               value2=sdr.min_max[1])

        self.min_dist_between = LabeledNumberInput("Minimum gap", nsn_factory.min_dist_between)

        self.dot_colour = LabeledInput("Colour", text=self.settings.default_object_colour,
                                       case_sensitive=False)
        self.dot_colour2 = LabeledInput("Colour 2", text="expyriment_orange", case_sensitive=False)

        ctrl = QVBoxLayout()
        ctrl.addWidget(self.btn_generate)
        ctrl.addSpacing(10)
        ctrl.addLayout(self.number.layout())
        ctrl.addLayout(self.number2.layout())

        ctrl.addWidget(heading("Dot"))
        ctrl.addLayout(self.item_diameter_mean.layout())
        ctrl.addLayout(self.item_diameter_range.layout())
        ctrl.addLayout(self.item_diameter_std.layout())
        ctrl.addLayout(self.dot_colour.layout())
        ctrl.addLayout(self.dot_colour2.layout())

        ctrl.addWidget(heading("Array"))
        ctrl.addLayout(self.target_area_radius.layout())
        ctrl.addLayout(self.min_dist_between.layout())
        ctrl.addStretch(1)

        # Add text field
        fields = QVBoxLayout()
        self._text_field = QTextBrowser(self)
        self._text_field.setFont(QFont("Courier New", 8))
        self._text_field.setReadOnly(True)
        self._text_field.setWordWrapMode(False)
        self._text_field.setVerticalScrollBar(QScrollBar())
        self._text_field.setHorizontalScrollBar(QScrollBar())
        self.picture_field = QLabel(self)
        self.resize_fields(400)
        fields.addWidget(self.picture_field)
        fields.addWidget(self._text_field)

        hlayout = QHBoxLayout()
        hlayout.addLayout(ctrl)
        hlayout.addLayout(fields)
        self.setLayout(hlayout)

        self.updateUI()

    def updateUI(self):
        self.dot_colour2.setVisible(self.settings.bicoloured.isChecked())
        self.number2.setVisible(self.settings.bicoloured.isChecked())

    def resize_fields(self, width, text_height=150, minium_text_width=300):
        self.picture_field.setFixedSize(width, width)
        if width < minium_text_width:
            width = minium_text_width
        self._text_field.setMinimumSize(width, text_height)

    def text_clear(self):
        self._text_field.clear()

    def text_out(self, txt):
        self._text_field.append(txt)

    def text_error_feedback(self, error):
        self.text_clear()
        self.text_out("ERROR\n=====\n")
        txt = '<span style=\" color: #ff0000;\">%s</span>' % str(error)
        self.text_out(txt)
        self.updateUI()




"""
Draw a random number from a beta dirstribution
"""

__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

from collections import OrderedDict
import numpy as np
import sys


def make_csv(xy, size_data_dict, attributes=None,
              array_hash=None, make_variable_names=True):
    """Not for User
    makes csv for Arrays with object of different size information
    size_data_dict: keys = variable names (e.g. width, height),
                    values vector of size data
    """
    rtn = ""
    if make_variable_names:
        if array_hash:
            rtn += "hash,"
        rtn += "x,y," + ",".join(size_data_dict.keys()) + ","
        if attributes is not None:
            rtn += "attribute,"
        rtn = rtn[:-1] + "\n"  # replace comma

    size_data = np.array(list(size_data_dict.values())).T
    if attributes is None:
        attribute_vector = [None] * len(xy)  # to have something to loop
    else:
        attribute_vector = attributes

    for pos, size, attr in zip(xy, size_data, attribute_vector):
        if array_hash:
            rtn += "{0},".format(array_hash)
        rtn += "{},{},".format(pos[0], pos[1])
        for s in size:
            rtn += "{},".format(s)
        if attributes is not None:
            rtn += "{},".format(attr)
        rtn = rtn[:-1] + "\n"  # replace comma
    return rtn


def join_dict_list(list_of_dicts):
    """make a dictionary of lists from a list of dictionaries"""
    rtn = OrderedDict()
    for d in list_of_dicts:
        for k, v in d.items():
            if k in rtn:
                rtn[k].append(v)
            else:
                rtn[k] = [v]
    return rtn


def dict_to_csv(dictionary, variable_names=False, dict_of_lists=False):
    d = OrderedDict(dictionary.items())
    rtn = ""
    if variable_names:
        rtn += ",".join(d.keys()) + "\n"

    if dict_of_lists:
        prop_np = np.asarray(list(d.values())).T  # list is requires in PY3
        for x in prop_np:
            rtn += ", ".join(map(lambda s: str(s), x)) + "\n"
    else:
        rtn += ",".join(map(lambda s: str(s), d.values())) + "\n"

    return rtn


def numpy_vector(x):
    """helper function:
    make an numpy vector from any element (list, arrays, and single data (str, numeric))
    """

    x = np.array(x)
    if x.ndim == 1:
        return x
    elif x.ndim == 0:
        return x.reshape(1)  # if one element only, make a array with one element
    else:
        return x.flatten()


def numpy_array_2d(two_d_data):
    """ensures well shaped to 2d numpy array"""
    rtn = np.asarray(two_d_data)
    if rtn.ndim == 1 and len(rtn) == 2:
        rtn = rtn.reshape((1, 2))
    if rtn.ndim != 2:
        raise ValueError("Bad shaped data: xy must be pair of xy-values or a list of xy-values")
    return rtn


def numpy_round2(array, decimals, int_type=np.int32):
    """rounds and changes to int type if decimals == 0"""
    array = np.round(array, decimals=decimals)
    if decimals == 0:
        return array.astype(int_type)
    else:
        return array


def is_all_equal(vector):
    # returns true if all elements are equal
    return len(np.unique(np.asarray(vector))) == 1


def dict_to_text(the_dict, col_a = 22, col_b = 14,
                 spacing_char=" "):
    rtn = None
    for k, v in the_dict.items():
        if rtn is None:
            key_str = "- " + k
            rtn = ""
        else:
            key_str = "  " + k

        value = "{}\n".format(v)
        len_col_b = col_b - len(value)
        if len_col_b<2:
            len_col_b = 2
        rtn += key_str + (spacing_char * (col_a - len(key_str))) + \
                         (" " * len_col_b) + value
    return rtn.rstrip()


def is_interactive_mode():
    """Returns if Python is running in interactive mode (such as IDLE or
    IPthon)

    Returns
    -------
    interactive_mode : boolean

    """
    #ipython?
    try:
        __IPYTHON__
        return True
    except NameError:
        pass

    is_idle = "idlelib.run" in sys.modules
    # ps2 is only defined in interactive mode
    return is_idle or hasattr(sys, "ps2")


def triu_nan(m, k=0):
    """helper function
    upper triangular but nan instead of zeros (as in numpy's original function,
    see docu numpy.triu)
    """
    return m + np.tril(np.full(m.shape, np.nan), k=k-1)


import sqlite3
from ..image._colour import Colour, ImageColours


class DotArraySQLDB(object):
    """DotArray DB (two tables: arrays, dots)"""

    def __init__(self, db_file, setup=False):
        self.db_file = db_file
        if setup:
            self.setup()

    def setup(self):
        conn = sqlite3.connect(self.db_file)

        cursor = conn.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = [v[0] for v in cursor.fetchall() if v[0] != "sqlite_sequence"]

        if 'ARRAYS' not in tables:
            conn.execute('''CREATE TABLE ARRAYS
                             (HASH varchar(32) PRIMARY KEY,
                             N   INT     NOT NULL,
                             TSA REAL NOT NULL,
                             ISA  REAL NOT NULL,
                             FA  REAL NOT NULL,
                             SPAR  REAL NOT NULL,
                             logSIZE REAL NOT NULL,
                             logSPACE REAL NOT NULL,
                             COV REAL NOT NULL,
                             COLOUR TEXT);''')

        if 'DOTS' not in tables:
            conn.execute('''CREATE TABLE DOTS
                     (HASH varchar(32) NOT NULL,
                     x REAL NOT NULL,
                     y REAL NOT NULL,
                     diameter REAL NOT NULL,
                     COLOUR TEXT);''')

        cursor.close()
        conn.close()

    def vacuum(self):
        conn = sqlite3.connect(self.db_file)
        conn.execute("VACUUM;")
        conn.close()

    def add_arrays(self, dot_arrays):

        if not isinstance(dot_arrays, (list, tuple)):
            dot_arrays = [dot_arrays]

        conn = sqlite3.connect(self.db_file)
        cur = conn.cursor()

        for da in dot_arrays:
            attributes = da.as_dict()['attributes']

            # assuming single colour dot arrays
            # TODO if multi color, add to dots color column. multi_colour = len(attributes)!=1
            try:
                colour = Colour(attributes[0])
            except:
                # if attribute is no colour, use default coolour
                colour = Colour(ImageColours.COL_DEFAULT_OBJECT)

            sql = "INSERT INTO ARRAYS (" + \
                  "HASH, N, TSA, ISA, FA, SPAR, logSIZE, logSPACE, COV, COLOUR" + \
                  ") \n VALUES\n" + \
                  "('{}',{},{},{},{},{},{},{},{},'{}');".format(
                          da.hash,
                          da._properties.numerosity,
                          da._properties.total_surface_area,
                          da._properties.average_surface_area,
                          da._properties.field_area,
                          da._properties.sparsity,
                          da._properties.log_size,
                          da._properties.log_spacing,
                          da._properties.converage,
                          colour.colour)
            cur.execute(sql)

            ## add dots
            sql = "INSERT INTO DOTS (HASH,x,y,diameter) \nVALUES"
            for xy, d in zip(da._xy, da._diameter):
                sql += "\n  ('{}', {}, {}, {}),".format(da.hash, xy[0], xy[1], d)
            sql = sql[:-1] + ";"
            cur.execute(sql)

        conn.commit()
        conn.close()

    def get_all_hashes(self):
        conn = sqlite3.connect(self.db_file)
        values = conn.execute("SELECT hash FROM ARRAYS;").fetchall()
        conn.close()
        return [x[0] for x in values]

    def get_array(self, hash):
        conn = sqlite3.connect(self.db_file)
        conn.row_factory = sqlite3.Row
        cur = conn.execute("SELECT * FROM ARRAYS WHERE hash ='{}';".format(hash))
        v = cur.fetchone()
        conn.close()
        return {k:v[k] for k in v.keys()}

    def get_dots(self, hash):
        # get ID
        conn = sqlite3.connect(self.db_file)
        values = conn.execute("SELECT x, y, diameter, colour FROM DOTS WHERE hash ='{}';".format(hash)).fetchall()
        conn.close()
        return values #




__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

from ..exceptions import NoSolutionError, NoAppearanceDefinedError
from .base_classes import ArrayParameter
from .._lib.misc import dict_to_text
from .dot_array import DotArray
from .rect_array import RectangleArray
from .appearance_sampler import AppearanceSampler


class NSNFactory(ArrayParameter, AppearanceSampler):

    def __init__(self, target_area_radius,
                 min_dist_between=None,
                 min_dist_area_boarder=None):
        """

        Parameters
        ----------
        target_area_radius
        min_dist_between
        min_dist_area_boarder
        """

        AppearanceSampler.__init__(self)
        ArrayParameter.__init__(self,
            target_area_radius=target_area_radius,
            min_dist_between=min_dist_between,
            min_dist_area_boarder=min_dist_area_boarder)

    def create_random_array(self, n_objects,
                            allow_overlapping=False,
                            occupied_space=None):
        """
        occupied_space is a dot array (used for multicolour dot array (join after)

        attribute is an array, arrays are assigned randomly.


        Parameters
        ----------
        n_objects
        allow_overlapping
        occupied_space

        Returns
        -------
        rtn : object array
        """
        if not self.is_appearance_set:
            raise NoAppearanceDefinedError("No appearance defined. Please use 'set_dot', 'set_rect_or' "
                                           "'set_appearance'")
        if self._distr_diameter is not None:
            # DotArray
            rtn = DotArray(target_area_radius=self.target_area_radius,
                                min_dist_between=self.min_dist_between,
                                min_dist_area_boarder=self.min_dist_area_boarder)

            for dot in self.sample(n=n_objects):
                try:
                    dot = rtn.get_free_position(ref_object=dot, in_neighborhood=False,
                                                occupied_space=occupied_space,
                                                allow_overlapping=allow_overlapping)
                except NoSolutionError as e:
                    raise NoSolutionError("Can't find a solution for {} items in this array".format(n_objects))
                rtn.add([dot])

        else:
            # RectArray
            rtn = RectangleArray(target_area_radius=self.target_area_radius,
                                      min_dist_between=self.min_dist_between,
                                      min_dist_area_boarder=self.min_dist_area_boarder)

            for rect in self.sample(n=n_objects):
                try:
                    rect = rtn.get_free_position(ref_object=rect, in_neighborhood=False,
                                                 occupied_space=occupied_space,
                                                 allow_overlapping=allow_overlapping)
                except NoSolutionError:
                    raise NoSolutionError("Can't find a solution for {} ".format(n_objects) +
                                          "items in this array.")

                rtn.add([rect])
        return rtn

    def create_incremental_random_array(self, n_objects,
                                        allow_overlapping=False):
        """

        Parameters
        ----------
        n_objects
        allow_overlapping

        Returns
        -------
        rtn : iterator of object arrays
        """
        previous = None
        for n in range(n_objects):
            current = self.create_random_array(n_objects=1,
                   allow_overlapping=allow_overlapping,
                   occupied_space=previous)
            if previous is not None:
                current.join(previous)
            previous = current
            yield current

    def as_dict(self):
        d = ArrayParameter.as_dict(self)
        d.update(AppearanceSampler.as_dict(self))
        return d

    def __str__(self):
        return dict_to_text(self.as_dict(), col_a=12, col_b=7)


# calculates visual properties of a dot array/ dot cloud

from collections import OrderedDict
from enum import IntFlag, auto
from math import log2

import numpy as np
from .. import _lib
from ._convex_hull import ConvexHull, ConvexHullPositions

class VisualPropertyFlag(IntFlag):

    AV_DOT_DIAMETER = auto()
    AV_SURFACE_AREA = auto()
    AV_PERIMETER = auto()
    AV_RECT_SIZE = auto()

    TOTAL_SURFACE_AREA = auto()
    TOTAL_PERIMETER = auto()
    SPARSITY = auto()
    FIELD_AREA = auto()
    FIELD_AREA_POSITIONS = auto()
    COVERAGE = auto()

    LOG_SPACING = auto()
    LOG_SIZE = auto()

    NUMEROSITY = auto()

    def is_dependent_from(self, other_property):
        """returns true if both properties are not independent"""
        return (self.is_size_property() and other_property.is_size_property()) or \
               (self.is_space_property() and other_property.is_space_property())

    def is_size_property(self):
        return self in (VisualPropertyFlag.LOG_SIZE,
                        VisualPropertyFlag.TOTAL_SURFACE_AREA,
                        VisualPropertyFlag.AV_DOT_DIAMETER,
                        VisualPropertyFlag.AV_SURFACE_AREA,
                        VisualPropertyFlag.AV_PERIMETER,
                        VisualPropertyFlag.TOTAL_PERIMETER)

    def is_space_property(self):
        return self in (VisualPropertyFlag.LOG_SPACING,
                        VisualPropertyFlag.SPARSITY,
                        VisualPropertyFlag.FIELD_AREA)

    def label(self):
        labels = {
            VisualPropertyFlag.NUMEROSITY: "Numerosity",
            VisualPropertyFlag.LOG_SIZE: "Log Size",
            VisualPropertyFlag.TOTAL_SURFACE_AREA: "Total surface area",
            VisualPropertyFlag.AV_DOT_DIAMETER: "Average dot diameter",
            VisualPropertyFlag.AV_SURFACE_AREA: "Average surface area",
            VisualPropertyFlag.AV_PERIMETER: "Average perimeter",
            VisualPropertyFlag.TOTAL_PERIMETER: "Total perimeter",
            VisualPropertyFlag.AV_RECT_SIZE: "Average Rectangle Size",
            VisualPropertyFlag.LOG_SPACING: "Log Spacing",
            VisualPropertyFlag.SPARSITY: "Sparsity",
            VisualPropertyFlag.FIELD_AREA: "Field area",
            VisualPropertyFlag.FIELD_AREA_POSITIONS: "Field area positions",
            VisualPropertyFlag.COVERAGE: "Coverage"}
        return labels[self]


class ArrayProperties(object):

    def __init__(self, object_array):
        # _lib or dot_cloud
        assert isinstance(object_array, (_lib.DotArray, _lib.RectangleArray,
                                         _lib.AttributeArray))
        self.oa = object_array
        self._convex_hull = None
        self._convex_hull_positions = None

    def reset(self):
        """reset to enforce recalculation of certain parameter """
        self._convex_hull = None
        self._convex_hull_positions = None

    @property
    def convex_hull(self):
        if self._convex_hull is None:
            self._convex_hull = ConvexHull(self.oa)
        return self._convex_hull

    @property
    def convex_hull_positions(self):
        if self._convex_hull_positions is None:
            self._convex_hull_positions = ConvexHullPositions(self.oa)
        return self._convex_hull_positions

    @property
    def average_dot_diameter(self):
        if not isinstance(self.oa, _lib.DotArray):
            return None
        elif self.numerosity == 0:
            return np.nan
        else:
            return np.mean(self.oa.diameter)

    @property
    def average_rectangle_size(self):
        if not isinstance(self.oa, _lib.RectangleArray):
            return None
        elif self.numerosity == 0:
            return np.array([np.nan, np.nan])
        else:
            return np.mean(self.oa.sizes, axis=0)

    @property
    def total_surface_area(self):
        return np.sum(self.oa.surface_areas)

    @property
    def average_surface_area(self):
        if self.numerosity == 0:
            return np.nan
        return np.mean(self.oa.surface_areas)

    @property
    def total_perimeter(self):
        return np.sum(self.oa.perimeter)

    @property
    def average_perimeter(self):
        if self.numerosity == 0:
            return np.nan
        return np.mean(self.oa.perimeter)

    @property
    def field_area_positions(self):
        return self.convex_hull_positions.field_area

    @property
    def numerosity(self):
        return len(self.oa._xy)

    @property
    def converage(self):
        """ percent coverage in the field area. It takes thus the object size
        into account. In contrast, the sparsity is only the ratio of field
        array and numerosity
        """
        try:
            return self.total_surface_area / self.field_area
        except ZeroDivisionError:
            return np.nan

    @property
    def log_size(self):
        try:
            return log2(self.total_surface_area) + \
                   log2(self.average_surface_area)
        except ValueError:
            return np.nan


    @property
    def log_spacing(self):
        try:
            return log2(self.field_area) + log2(self.sparsity)
        except ValueError:
            return np.nan

    @property
    def sparsity(self):
        try:
            return self.field_area / self.numerosity
        except ZeroDivisionError:
            return np.nan

    @property
    def field_area(self):
        return self.convex_hull.field_area

    def get(self, property_flag):
        """returns a visual property"""

        assert isinstance(property_flag, VisualPropertyFlag)

        # Adapt
        if property_flag == VisualPropertyFlag.AV_DOT_DIAMETER:
            return self.average_dot_diameter

        elif property_flag == VisualPropertyFlag.AV_RECT_SIZE:
            return self.average_rectangle_size

        elif property_flag == VisualPropertyFlag.AV_PERIMETER:
            return self.average_perimeter

        elif property_flag == VisualPropertyFlag.TOTAL_PERIMETER:
            return self.total_perimeter

        elif property_flag == VisualPropertyFlag.AV_SURFACE_AREA:
            return self.average_surface_area

        elif property_flag == VisualPropertyFlag.TOTAL_SURFACE_AREA:
            return self.total_surface_area

        elif property_flag == VisualPropertyFlag.LOG_SIZE:
            return self.log_size

        elif property_flag == VisualPropertyFlag.LOG_SPACING:
            return self.log_spacing

        elif property_flag == VisualPropertyFlag.SPARSITY:
            return self.sparsity

        elif property_flag == VisualPropertyFlag.FIELD_AREA:
            return self.field_area

        elif property_flag == VisualPropertyFlag.FIELD_AREA_POSITIONS:
            return self.field_area_positions

        elif property_flag == VisualPropertyFlag.COVERAGE:
            return self.converage

        else:
            raise ValueError("{} is a unknown visual feature".format(property_flag))

    def as_dict(self):
        """ordered dictionary with the most important visual properties"""
        rtn = [("Hash", self.oa.hash),
               ("Numerosity", self.numerosity),
               ("?", None),  # placeholder
               (VisualPropertyFlag.AV_PERIMETER.label(), self.average_perimeter),
               (VisualPropertyFlag.AV_SURFACE_AREA.label(), self.average_surface_area),
               (VisualPropertyFlag.TOTAL_PERIMETER.label(), self.total_perimeter),
               (VisualPropertyFlag.TOTAL_SURFACE_AREA.label(), self.total_surface_area),
               (VisualPropertyFlag.FIELD_AREA.label(), self.field_area),
               (VisualPropertyFlag.SPARSITY.label(), self.sparsity),
               (VisualPropertyFlag.COVERAGE.label(), self.converage),
               (VisualPropertyFlag.LOG_SIZE.label(), self.log_size),
               (VisualPropertyFlag.LOG_SPACING.label(), self.log_spacing)]

        if isinstance(self.oa, _lib.DotArray):
            rtn[2] = (VisualPropertyFlag.AV_DOT_DIAMETER.label(), self.average_dot_diameter)
        elif isinstance(self.oa, _lib.RectangleArray):
            rtn[2] = (VisualPropertyFlag.AV_RECT_SIZE.label(), self.average_rectangle_size)
        else:
            rtn.pop(2)
        return OrderedDict(rtn)

    def __str__(self):
        return self.as_text()

    def as_text(self, with_hash=True, extended_format=False, spacing_char="."):
        if extended_format:
            rtn = None
            for k, v in self.as_dict().items():
                if rtn is None:
                    if with_hash:
                        rtn = "- {}: {}\n".format(k, v)
                    else:
                        rtn = ""
                else:
                    if rtn == "":
                        name = "- " + k
                    else:
                        name = "  " + k
                    try:
                        value = "{0:.2f}\n".format(v)  # try rounding
                    except:
                        value = "{}\n".format(v)

                    n_space = 14 - len(value)
                    if n_space < 2:
                        n_space = 2
                    rtn += name + (spacing_char * (24 - len(name))) + (" " * n_space) + value
        else:
            if with_hash:
                rtn = "ID: {} ".format(self.oa.hash)
            else:
                rtn = ""
            rtn += "N: {}, TSA: {}, ISA: {}, FA: {}, SPAR: {:.3f}, logSIZE: " \
                   "{:.2f}, logSPACE: {:.2f} COV: {:.2f}".format(
                self.numerosity,
                int(self.total_surface_area),
                int(self.average_surface_area),
                int(self.field_area),
                self.sparsity,
                self.log_size,
                self.log_spacing,
                self.converage)
        return rtn.rstrip()



"""
Dot Array
"""

__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import numpy as np

from .base_classes import ABCObjectArray
from .._lib import misc, geometry
from .shapes import Dot, Point
from . import array_tools

# TODO: How to deal with rounding? Is saving to precises? Suggestion:
#  introduction precision parameter that is used by as_dict and get_csv and
#  hash


class DotArray(ABCObjectArray):
    """Numpy Position list for optimized for numpy calculations


    Position + diameter
    """

    def __init__(self,
                 target_area_radius,
                 min_dist_between=None,
                 min_dist_area_boarder=None,
                 xy = None,
                 diameter = None,
                 attributes = None):
        """Dot array is restricted to a certain area, it has a target area
        and a minimum gap.

        This properties allows shuffling free position and adapting
        properties.
        """
        super().__init__(xy=xy, attributes=attributes,
                         target_area_radius=target_area_radius,
                         min_dist_between=min_dist_between,
                         min_dist_area_boarder=min_dist_area_boarder)
        if diameter is None:
            self._diameter = np.array([])
        else:
            self._diameter = misc.numpy_vector(diameter)
        if self._xy.shape[0] != len(self._diameter):
            raise ValueError("Bad shaped data: " +
                             u"xy has not the same length as item_diameter")

    def add(self, dots):
        """append one dot or list of dots"""
        try:
            dots = list(dots)
        except TypeError:
            dots = [dots]
        for d in dots:
            assert isinstance(d, Dot)
            self._append_xy_attribute(xy=d.xy, attributes=d.attribute)
            self._diameter = np.append(self._diameter, d.diameter)
        self.properties.reset()

    @property
    def diameter(self):
        return self._diameter

    @property
    def surface_areas(self):
        # a = pi r**2 = pi d**2 / 4
        return np.pi * (self._diameter ** 2) / 4.0

    @property
    def perimeter(self):
        return np.pi * self._diameter

    def mod_round_values(self, decimals=0, int_type=np.int32):
        """Round values of the array."""

        if decimals is None:
            return
        self._xy = misc.numpy_round2(self._xy, decimals=decimals,
                                     int_type=int_type)
        self._diameter = misc.numpy_round2(self._diameter, decimals=decimals,
                                           int_type=int_type)

    def as_dict(self):
        """
        """
        d = super().as_dict()
        d.update({"diameter": self._diameter.tolist()})
        return d

    def read_from_dict(self, the_dict):
        """read Dot collection from dict"""
        super().read_from_dict(the_dict)
        self._diameter = np.array(the_dict["diameter"])

    def clear(self):
        super().clear()
        self._diameter = np.array([])

    def delete(self, index):
        super().delete(index)
        self._diameter = np.delete(self._diameter, index)

    def copy(self, indices=None, deepcopy=True):
        """returns a (deep) copy of the dot array.

        It allows to copy a subset of dot only.

        """

        if len(self._xy) == 0:
            return DotArray(target_area_radius=self.target_area_radius,
                            min_dist_between=self.min_dist_between,
                            min_dist_area_boarder=self.min_dist_area_boarder)

        if indices is None:
            indices = list(range(len(self._xy) ))

        if deepcopy:
            return DotArray(target_area_radius=self.target_area_radius,
                            min_dist_between=self.min_dist_between,
                            min_dist_area_boarder = self.min_dist_area_boarder,
                            xy=self._xy[indices, :].copy(),
                            diameter=self._diameter[indices].copy(),
                            attributes=self._attributes[indices].copy())
        else:
            return DotArray(target_area_radius=self.target_area_radius,
                            min_dist_between=self.min_dist_between,
                            min_dist_area_boarder = self.min_dist_area_boarder,
                            xy=self._xy[indices, :],
                            diameter=self._diameter[indices],
                            attributes=self._attributes[indices])

    def get_distances(self, dot):
        """Distances toward a single dot
        negative numbers indicate overlap

        Returns
        -------
        distances : numpy array of distances
        """
        assert isinstance(dot, Dot)
        if len(self._xy) == 0:
            return np.array([])
        else:
            rtn = np.hypot(self._xy[:, 0] - dot.x, self._xy[:, 1] - dot.y) - \
                   ((self._diameter + dot.diameter) / 2.0)
            return rtn

    def iter_objects(self, indices=None):
        """iterate over all or a part of the objects

        Parameters
        ----------
        indices: int or interable of integer

        Notes
        -----
        To iterate all object you might all use the class iterator __iter__:
        >>> for obj in my_array:
        >>>    print(obj)
        """

        if isinstance(indices, (int, np.integer)):
            yield Dot(xy=self._xy[indices, :],
                      diameter=self._diameter[indices],
                      attribute=self._attributes[indices])
        else:
            if indices is None:
                data = zip(self._xy, self._diameter, self._attributes)
            else:
                data = zip(self._xy[indices, :],  self._diameter[indices],
                                    self._attributes[indices])
            for xy, dia, att in data:
                yield Dot(xy=xy, diameter=dia, attribute=att)

    def find_objects(self, diameter=None, attribute=None):
        """returns indices of found objects
        """
        rtn = []
        for i in range(len(self._diameter)):
            if (diameter is not None and self._diameter[i] != diameter) or \
                    (attribute is not None and self._attributes[i] != attribute):
                continue
            rtn.append(i)
        return rtn

    def csv(self, variable_names=True, hash_column=False,
            attribute_column=False):
        """Return the dot array as csv text

        Parameter
        ---------
        variable_names : bool, optional
            if True variable name will be printed in the first line
        """
        size_dict = {"diameter": self._diameter}
        if attribute_column:
            attr = self.attributes
        else:
            attr = None
        if hash_column:
            array_hash = self.hash
        else:
            array_hash = None
        return misc.make_csv(xy=self._xy,
                             size_data_dict=size_dict,
                             attributes=attr, array_hash=array_hash,
                             make_variable_names=variable_names)


__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import numpy as _np
import svgwrite as _svg
from .._lib.geometry import cartesian2image_coordinates as _c2i_coord
from .. import _lib
from . import _array_draw


def create(object_array, colours, filename):
    """SVG image/file, vector image format

    Parameters
    ----------
    object_array
    colours
    filename

    Returns
    -------

    """
    return _SVGDraw().create_image(object_array=object_array, colours=colours,
                                   filename=filename)


class _SVGDraw(_array_draw.ArrayDraw):
    # scaling not used, because vector format is scale independent.

    @staticmethod
    def get_squared_image(image_width, background_colour, **kwargs):
        """"""
        px = "{}px".format(image_width)
        return _svg.Drawing(size=(px, px), filename=kwargs['filename'])

    @staticmethod
    def scale_image(image, scaling_factor):
        """"""
        return image # not used

    @staticmethod
    def draw_shape(image, shape, opacity, scaling_factor):
        """"""
        assert isinstance(image, _svg.Drawing)
        width = int(image.attribs['width'][:-2]) # string "300px" --> 300
        shape.xy = _c2i_coord(_np.asarray(shape.xy), width).tolist()
        attr = shape.get_attribute_object()

        if isinstance(attr, _lib.PictureFile):
            raise RuntimeError("Pictures are not supported for SVG file.")

        if isinstance(shape, _lib.Dot):
            image.add(image.circle(center=shape.xy,
                                       r=shape.diameter / 2,
                                       # stroke_width="0", stroke="black",
                                       fill=attr.colour,
                                       opacity=opacity))
        elif isinstance(shape, _lib.Rectangle):
            image.add(image.rect(insert=(shape.left, shape.bottom),
                                     size=shape.size,
                                     fill=attr.colour,
                                     opacity=opacity))
        else:
            raise NotImplementedError("Shape {} NOT YET IMPLEMENTED".format(type(shape)))

    @staticmethod
    def draw_convex_hull(image, points, convex_hull_colour, opacity,
                         scaling_factor):
        """"""
        width = int(image.attribs['width'][:-2]) # string "300px" --> 300
        points = _c2i_coord(points, width)

        last = None
        for p in _np.append(points, [points[0]], axis=0):
            if last is not None:
                l = image.line(start=last, end=p).stroke(
                    width=1, color=convex_hull_colour.colour, opacity=opacity)
                image.add(l)
            last = p



__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import numpy as _np
from matplotlib import pyplot as _plt
from .. import _lib
from . import _array_draw


def create(object_array, colours=None, dpi=100):
    """Matplotlib figure

    Parameters
    ----------
    dpi
    object_array
    colours

    Returns
    -------

    """
    return _MatplotlibDraw().create_image(object_array=object_array, colours=colours,
                                          dpi=dpi)


class _MatplotlibDraw(_array_draw.ArrayDraw):

    @staticmethod
    def get_squared_image(image_width, background_colour,  **kwargs):
        dpi = kwargs["dpi"]
        figure = _plt.figure(figsize=_np.array([image_width, image_width]) / dpi,
                             dpi=dpi)
        if background_colour is None:
            figure.patch.set_facecolor((0, 0, 0, 0))
        else:
            figure.patch.set_facecolor(background_colour)
        axes = _plt.Axes(figure, [0., 0., 1, 1])
        axes.set_aspect('equal')  # squared
        axes.set_axis_off()
        r = image_width / 2
        axes.set(xlim=[-1 * r, r], ylim=[-1 * r, r])
        figure.add_axes(axes)
        return figure

    @staticmethod
    def scale_image(figure, scaling_factor):
        # not used
        pass

    @staticmethod
    def draw_shape(img, shape, opacity, scaling_factor):
        attr = shape.get_attribute_object()

        if isinstance(attr, _lib.PictureFile):
            raise RuntimeError("Pictures are not supported for matplotlib files.")

        if isinstance(shape, _lib.Dot):
            r = shape.diameter / 2
            plt_shape = _plt.Circle(xy=shape.xy, radius=r, color=attr.colour,
                                    lw=0)
        elif isinstance(shape, _lib.Rectangle):
            xy = (shape.left, shape.bottom)
            plt_shape = _plt.Rectangle(xy=xy,
                                       width=shape.width,
                                       height=shape.height,
                                       color=attr.colour, lw=0)
        else:
            raise NotImplementedError("Shape {} NOT YET IMPLEMENTED".format(type(shape)))

        plt_shape.set_alpha(opacity)
        img.axes[0].add_artist(plt_shape)


    @staticmethod
    def draw_convex_hull(img, points, convex_hull_colour, opacity,
                         scaling_factor):
        hull = _np.append(points, [points[0]], axis=0)
        for i in range(1, hull.shape[0]):
            line = _plt.Line2D(xdata=hull[i - 1:i + 1, 0],
                               ydata=hull[i - 1:i + 1, 1],
                               linewidth=1, color=convex_hull_colour.colour)
            line.set_alpha(opacity)
            img.axes[0].add_artist(line)





import numpy as _np
from math import log2 as _log2
from .._lib import geometry as _geometry
from .._lib import rng as _rng
from .. import _lib
from .. import constants
from ..exceptions import NoSolutionError as _NoSolutionError
from ._properties import VisualPropertyFlag as _flags


def change_fit_settings(default_spacing_precision=None,
                        default_fa2ta_ratio=None):
    """Changing class settings of property fitting.

    This changes the settings of the property fitting.


    Parameters
    ----------
    default_spacing_precision
    default_fa2ta_ratio

    Returns
    -------

    """

    if default_spacing_precision is not None:
        constants.DEFAULT_FIT_SPACING_PRECISION = float(default_spacing_precision)
    if default_fa2ta_ratio is not None:
        constants.DEFAULT_FIT_FA2TA_RATIO = float(default_fa2ta_ratio)

#FIXME coverage for all


def numerosity(object_array, value, keep_convex_hull=False):
    """

    """
    _lib._check_object_array(object_array)

    # make a copy for the deviant
    if value <= 0:
        object_array.clear()
    else:
        # add or remove random dots
        change_numerosity = value - object_array._properties.numerosity
        for _ in range(abs(change_numerosity)):
            if change_numerosity < 0:
                # remove dots
                if keep_convex_hull:
                    # find a random object that is not in convex hull
                    delete_id = None
                    ch = object_array._properties.convex_hull.object_indices_unique
                    rnd_seq = list(range(0, object_array._properties.numerosity))
                    _rng.generator.shuffle(rnd_seq)
                    for x in rnd_seq:
                        if x not in ch:
                            delete_id = x
                            break
                    if delete_id is None:
                        raise _NoSolutionError("Can't increase numeroisty, while keeping field area.")
                else:
                    delete_id = _rng.generator.integers(0, object_array._properties.numerosity)

                object_array.delete(delete_id)

            else:
                # add dot: copy a random dot
                clone_id = _rng.generator.integers(0, object_array._properties.numerosity)
                rnd_object = next(object_array.iter_objects(clone_id))
                try:
                    rnd_object = object_array.get_free_position(
                        ref_object=rnd_object, allow_overlapping=False,
                        inside_convex_hull=keep_convex_hull)
                except _NoSolutionError:
                    # no free position
                    raise _NoSolutionError("Can't increase numerosity. No free position found.")

                object_array.add([rnd_object])

    return object_array


def average_diameter(dot_array, value):
    # changes diameter
    if not isinstance(dot_array, _lib.DotArray):
        raise TypeError("Adapting diameter is not possible for {}.".format(
            type(dot_array).__name__))
    scale = value / dot_array.properties.average_dot_diameter
    dot_array._diameter = dot_array.diameter * scale
    dot_array.properties.reset()
    return dot_array


def average_rectangle_size(rect_array, value):
    # changes diameter
    if not isinstance(rect_array, _lib.RectangleArray):
        raise TypeError("Adapting rectangle size is not possible for {}.".format(
            type(rect_array).__name__))
    try:
        width, height = value
    except TypeError:
        raise TypeError("Value ({}) has to tuple of 2 numerical (width, height).".format(
            value))

    scale = _np.array([width / rect_array.properties.average_rectangle_size[0],
                       height / rect_array.properties.average_rectangle_size[1]])
    rect_array._sizes = rect_array._sizes * scale
    rect_array.properties.reset()
    return rect_array


def total_surface_area(object_array, value):
    # changes diameter
    _lib._check_object_array(object_array)
    a_scale = value / object_array.properties.total_surface_area
    if isinstance(object_array, _lib.DotArray):
        object_array._diameter = _np.sqrt(
            object_array.surface_areas * a_scale) * 2 / _np.sqrt(
                    _np.pi)  # d=sqrt(4a/pi) = sqrt(a)*2/sqrt(pi)
    else: # rect
        object_array._sizes = object_array._sizes * _np.sqrt(a_scale)

    object_array.properties.reset()
    return object_array


def field_area(object_array, value, precision=None):
    """changes the convex hull area to a desired size with certain precision

    uses scaling radial positions if field area has to be increased
    uses replacement of outer points (and later re-scaling)

    iterative method can takes some time.
    """

    _lib._check_object_array(object_array)
    if precision is None:
        precision = constants.DEFAULT_FIT_SPACING_PRECISION

    if object_array.properties.field_area is _np.nan:
        return  object_array # not defined
    else:
        return _scale_field_area(object_array, value=value,
                                 precision=precision)


def _scale_field_area(object_array, value, precision):
    """change the convex hull area to a desired size by scale the polar
    positions with certain precision

    iterative method can takes some time.

    Note: see doc string `field_area`
    """
    _lib._check_object_array(object_array)
    current = object_array.properties.field_area

    if current is None:
        return  # not defined

    scale = 1  # find good scale
    step = 0.1
    if value < current:  # current too larger
        step *= -1

    # centered points
    old_center = object_array.get_center_of_field_area()
    object_array._xy = object_array.xy - old_center
    centered_polar = _geometry.cartesian2polar(object_array.xy)

    # iteratively determine scale
    while abs(current - value) > precision:

        scale += step

        object_array._xy = _geometry.polar2cartesian(centered_polar * [scale, 1])
        object_array.properties.reset()  # required at this point to recalc convex hull
        current = object_array.properties.field_area

        if (current < value and step < 0) or \
                (current > value and step > 0):
            step *= -0.2  # change direction and finer grain

    object_array._xy = object_array.xy + old_center
    object_array.properties.reset()
    return object_array


def coverage(object_array, value,
             precision=None,
             FA2TA_ratio=None):

    # FIXME check drifting outwards if extra space is small and adapt_FA2TA_ratio=1
    # FIXME when to realign, realignment changes field_area!
    """this function changes the area and remixes to get a desired density
    precision in percent between 1 < 0

    ratio_area_convex_hull_adaptation:
        ratio of adaptation via area or via convex_hull (between 0 and 1)

    """
    _lib._check_object_array(object_array)

    print("WARNING: _adapt_coverage is a experimental ")
    # dens = convex_hull_area / total_surface_area
    if FA2TA_ratio is None:
        FA2TA_ratio = constants.DEFAULT_FIT_FA2TA_RATIO
    elif FA2TA_ratio < 0 or FA2TA_ratio > 1:
        FA2TA_ratio = 0.5
    if precision is None:
        precision = constants.DEFAULT_FIT_SPACING_PRECISION

    total_area_change100 = (value * object_array.properties.field_area) - \
                           object_array.properties.total_surface_area
    d_change_total_area = total_area_change100 * (1 - FA2TA_ratio)
    if abs(d_change_total_area) > 0:
        total_surface_area(object_array.properties.total_surface_area + \
                           d_change_total_area)

    return field_area(object_array.properties.total_surface_area / value,
               precision=precision)


def average_perimeter(object_array, value):
    _lib._check_object_array(object_array)
    total_peri = value * object_array.properties.numerosity
    return total_perimeter(object_array, total_peri)


def total_perimeter(object_array, value):
    if isinstance(object_array, _lib.DotArray):
        tmp = value / (object_array.properties.numerosity * _np.pi)
        return average_diameter(object_array, tmp)
    elif isinstance(object_array, _lib.RectangleArray):
        scale = value / object_array.properties.total_perimeter
        new_size = object_array.properties.average_rectangle_size * scale
        return average_rectangle_size(object_array, new_size)
    else:
        _lib._check_object_array(object_array)


def average_surface_area(object_array, value):
    _lib._check_object_array(object_array)
    ta = object_array.properties.numerosity * value
    return total_surface_area(object_array, ta)


def log_spacing(object_array, value, precision=None):
    _lib._check_object_array(object_array)
    logfa = 0.5 * value + 0.5 * _log2(
        object_array.properties.numerosity)
    return field_area(object_array, value=2 ** logfa, precision=precision)


def log_size(object_array, value):
    _lib._check_object_array(object_array)
    logtsa = 0.5 * value + 0.5 * _log2(object_array.properties.numerosity)
    return total_surface_area(object_array, 2 ** logtsa)


def sparcity(object_array, value, precision=None):
    _lib._check_object_array(object_array)
    return field_area(object_array, value=value * object_array.properties.numerosity,
                      precision=precision)


def visual_property(object_array, property_flag, value):
    """
    adapt_properties: continuous property or list of continuous properties
    several properties to be adapted
    if adapt dot array is specified, array will be adapt to adapt_dot_array, otherwise
    the values defined in adapt_properties is used.
    some adapting requires realignement to avoid overlaps. However,
    realigment might result in a different field area. Thus, realign after
    adapting for  Size parameter and realign before adapting space
    parameter.

    """

    # type check
    if not isinstance(property_flag, _flags):
        raise ValueError("{} is not a visual feature.".format(property_flag))

    # Adapt
    if property_flag == _flags.AV_DOT_DIAMETER:
        return average_diameter(object_array, value=value)

    elif property_flag == _flags.NUMEROSITY:
        return numerosity(object_array, value=value)

    elif property_flag == _flags.AV_PERIMETER:
        return average_perimeter(object_array, value=value)

    elif property_flag == _flags.TOTAL_PERIMETER:
        return total_perimeter(object_array, value=value)

    elif property_flag == _flags.AV_SURFACE_AREA:
        return average_surface_area(object_array, value=value)

    elif property_flag == _flags.TOTAL_SURFACE_AREA:
        return total_surface_area(object_array, value=value)

    elif property_flag == _flags.LOG_SIZE:
        return log_size(object_array, value=value)

    elif property_flag == _flags.LOG_SPACING:
        return log_spacing(object_array, value=value)

    elif property_flag == _flags.SPARSITY:
        return sparcity(object_array, value=value)

    elif property_flag == _flags.FIELD_AREA:
        return field_area(object_array, value=value)

    elif property_flag == _flags.COVERAGE:
        return coverage(object_array, value=value)

    else:
        raise NotImplementedError("Not implemented for {}".format(
            property_flag.label()))

# TODO "visual test" (eye inspection) of fitting rect arrays



"""
PyNSN package

Creating Non-Symbolic Number Displays
"""

__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'
__version__ = '0.11.10'


from sys import version_info as _python_version_info

if not (_python_version_info[0] >= 3 and _python_version_info[1] >= 6):
    raise RuntimeError("PyNSN {0} ".format(__version__) +
                       "is not compatible with Python {0}.{1}. ".format(
                           _python_version_info[0],
                           _python_version_info[1]) +
                       "Please use Python 3.6 or later.")

from ._lib import Point, Dot, Rectangle, \
    DotArray, RectangleArray, ArrayParameter, \
    PictureFile, NSNFactory
from .image._colour import Colour, ImageColours
from ._lib.rng import init_random_generator

from . import exceptions
from . import visual_properties
from . import distributions


def _print_version_info():
    from ._lib.misc import is_interactive_mode
    if is_interactive_mode():
        print("PyNSN {}".format(__version__))

_print_version_info()



"""
"""
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QCheckBox, \
    QDialogButtonBox, QComboBox, QHBoxLayout

from . import misc
from .. import constants
from ..visual_properties import flags, fit



class AdaptPropertyDialog(QDialog):

    def __init__(self, parent, properties):
        super(AdaptPropertyDialog, self).__init__(parent)

        self.setWindowTitle("Adapt Dot Array Property")
        self.properties = properties
        self._selection = None
        self.comboBox = QComboBox(self)
        for feat in flags:
            self.comboBox.addItem(feat.label())

        self.comboBox.activated[str].connect(self.choice)

        self._num_input = misc.NumberInput(width_edit=150, value=0)
        self.choice(flags.AV_DOT_DIAMETER)



        vlayout = QVBoxLayout(self)
        hlayout = QHBoxLayout()

        hlayout.addWidget(self.comboBox)
        hlayout.addWidget(self._num_input.edit)

        # OK and Cancel buttons
        buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
            Qt.Horizontal, self)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

        vlayout.addLayout(hlayout)
        vlayout.addWidget(buttons)

    def choice(self, selection):
        for feat in flags:
            if selection == feat:
                self._num_input.value = self.properties[feat.label()]
                self._selection = feat


    def current_state(self):
        return self._selection, self._num_input.value

    @staticmethod
    def get_response(parent, prop):
        """return the property to be adapted"""

        dialog = AdaptPropertyDialog(parent, prop)
        result = dialog.exec_()
        if result == QDialog.Accepted:
            return dialog.current_state()
        else:
            return None


class SettingsDialog(QDialog):

    def __init__(self, parent, image_colours):
        super(SettingsDialog, self).__init__(parent)


        self.setWindowTitle("Dot Array Property")

        self.rounding_decimals = misc.LabeledNumberInput(
                            label="Rounding decimals",
                            value=0, integer_only = True,
                            min = 0, max = 8)

        self.colour_area = misc.LabeledInput("Target Area",
                                             text=image_colours.target_area.colour,
                                             case_sensitive=False)
        self.colour_background = misc.LabeledInput("Background",
                                                   text=image_colours.background.colour,
                                                   case_sensitive=False)
        self.colour_convex_hull_positions = misc.LabeledInput("Colour field position",
                                                              text=image_colours.field_area_positions.colour,
                                                              case_sensitive=False)
        self.colour_convex_hull_dots = misc.LabeledInput("Colour field area",
                                                         text=image_colours.field_area.colour,
                                                         case_sensitive=False)
        self.antialiasing = QCheckBox("Antialiasing")
        self.antialiasing.setChecked(True)

        self.bicoloured = QCheckBox("bicoloured")
        self.bicoloured.setChecked(False)

        self.default_object_colour = image_colours.default_object_colour

        vlayout = QVBoxLayout()
        vlayout.addLayout(self.rounding_decimals.layout())

        vlayout.addWidget(misc.heading("Colour"))
        vlayout.addLayout(self.colour_area.layout())
        vlayout.addLayout(self.colour_background.layout())

        vlayout.addWidget(misc.heading("Convex hull"))
        vlayout.addLayout(self.colour_convex_hull_positions.layout())
        vlayout.addLayout(self.colour_convex_hull_dots.layout())
        vlayout.addSpacing(10)
        vlayout.addWidget(self.antialiasing)
        vlayout.addWidget(self.bicoloured)
        vlayout.addStretch(1)

        # OK and Cancel buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Ok, Qt.Horizontal, self)
        buttons.accepted.connect(self.accept)

        vlayout.addWidget(buttons)
        self.setLayout(vlayout)


class SequenceDialog(QDialog):
    extra_space = 50
    sequence_range = [10, 100]
    spacing_precision = constants.DEFAULT_FIT_SPACING_PRECISION
    adapt_FA2TA_ratio = constants.DEFAULT_FIT_FA2TA_RATIO


    def __init__(self, parent):

        super(SequenceDialog, self).__init__(parent)

        self.adapt_methods = []

        self.setWindowTitle("Sequence Dialog")

        self.adapt_diameter = QCheckBox(flags.AV_DOT_DIAMETER.label())
        self.adapt_av_perimeter= QCheckBox(flags.AV_PERIMETER.label())
        self.adapt_av_area = QCheckBox(flags.AV_SURFACE_AREA.label())
        self.adapt_area = QCheckBox(flags.TOTAL_SURFACE_AREA.label())
        self.adapt_total_perimeter = QCheckBox(flags.TOTAL_PERIMETER.label())
        self.adapt_coverage = QCheckBox(flags.COVERAGE.label())
        self.adapt_sparsity = QCheckBox(flags.SPARSITY.label())

        self.adapt_convex_hull = QCheckBox(flags.FIELD_AREA.label())
        self.adapt_size = QCheckBox(flags.LOG_SIZE.label())
        self.adapt_spacing = QCheckBox(flags.LOG_SPACING.label())
        self.adapt_spacing_presision = misc.LabeledNumberInput("Convex_hull presision",
                                                               value=SequenceDialog.spacing_precision,
                                                               integer_only=False)
        self.adapt_fa2ta = misc.LabeledNumberInput("Ratio convex_hull/area",
                                                   value=SequenceDialog.adapt_FA2TA_ratio,
                                                   integer_only=False, min=0, max=1)
        self.adapt_range = misc.LabeledNumberInputTwoValues("Sequence Range",
                                                            value1=SequenceDialog.sequence_range[0],
                                                            value2=SequenceDialog.sequence_range[1])
        self.adapt_extra_space = misc.LabeledNumberInput("Extra space",
                                                         value=SequenceDialog.extra_space,
                                                         integer_only=True, min=0)

        self.adapt_area.toggled.connect(self.ui_update)
        self.adapt_convex_hull.toggled.connect(self.ui_update)
        self.adapt_diameter.toggled.connect(self.ui_update)
        self.adapt_av_area.toggled.connect(self.ui_update)
        self.adapt_av_perimeter.toggled.connect(self.ui_update)
        self.adapt_total_perimeter.toggled.connect(self.ui_update)
        self.adapt_size.toggled.connect(self.ui_update)
        self.adapt_spacing.toggled.connect(self.ui_update)
        self.adapt_coverage.toggled.connect(self.ui_update)
        self.adapt_sparsity.toggled.connect(self.ui_update)
        self.adapt_spacing_presision.edit.editingFinished.connect(self.ui_update)
        self.adapt_fa2ta.edit.editingFinished.connect(self.ui_update)

        # OK and Cancel buttons
        buttons = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
            Qt.Horizontal, self)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)

        vlayout = QVBoxLayout()
        vlayout.addWidget(misc.heading("Adapting"))
        vlayout.addWidget(self.adapt_diameter)
        vlayout.addWidget(self.adapt_av_perimeter)
        vlayout.addWidget(self.adapt_av_area)
        vlayout.addWidget(self.adapt_area)
        vlayout.addWidget(self.adapt_total_perimeter)
        vlayout.addWidget(self.adapt_convex_hull)
        vlayout.addWidget(self.adapt_coverage)
        vlayout.addWidget(self.adapt_sparsity)
        vlayout.addSpacing(10)
        vlayout.addWidget(self.adapt_size)
        vlayout.addWidget(self.adapt_spacing)
        vlayout.addSpacing(10)
        vlayout.addWidget(misc.heading("Adapting parameter"))
        vlayout.addLayout(self.adapt_spacing_presision.layout())
        vlayout.addLayout(self.adapt_fa2ta.layout())
        vlayout.addLayout(self.adapt_range.layout())
        vlayout.addSpacing(10)
        vlayout.addLayout(self.adapt_extra_space.layout())
        vlayout.addSpacing(10)

        vlayout.addWidget(buttons)
        self.setLayout(vlayout)
        self.ui_update()

    def ui_update(self):
        # get methods
        selected = []
        all = [flags.AV_DOT_DIAMETER]
        if self.adapt_diameter.isChecked():
            selected.append(all[-1])

        all.append(flags.AV_SURFACE_AREA)
        if self.adapt_av_area.isChecked():
            selected.append(all[-1])

        all.append(flags.AV_PERIMETER)
        if self.adapt_av_perimeter.isChecked():
            selected.append(all[-1])

        all.append(flags.TOTAL_PERIMETER)
        if self.adapt_total_perimeter.isChecked():
            selected.append(all[-1])

        all.append(flags.TOTAL_SURFACE_AREA)
        if self.adapt_area.isChecked():
            selected.append(all[-1])

        all.append(flags.FIELD_AREA)
        if self.adapt_convex_hull.isChecked():
            selected.append(all[-1])

        all.append(flags.COVERAGE)
        if self.adapt_coverage.isChecked():
            selected.append(all[-1])

        all.append(flags.SPARSITY)
        if self.adapt_sparsity.isChecked():
            selected.append(all[-1])

        all.append(flags.LOG_SIZE)
        if self.adapt_size.isChecked():
            selected.append(all[-1])

        all.append(flags.LOG_SPACING)
        if self.adapt_spacing.isChecked():
            selected.append(all[-1])

        self.adapt_diameter.setEnabled(True)
        self.adapt_av_perimeter.setEnabled(True)
        self.adapt_av_area.setEnabled(True)
        self.adapt_area.setEnabled(True)
        self.adapt_total_perimeter.setEnabled(True)
        self.adapt_coverage.setEnabled(True)
        self.adapt_sparsity.setEnabled(True)
        self.adapt_convex_hull.setEnabled(True)
        self.adapt_spacing.setEnabled(True)
        self.adapt_size.setEnabled(True)

        for x in all:
            if x not in selected:
                # test dependency of non-selected item, x, from any selected
                check = [s.is_dependent_from(x) for s in selected]
                if sum(check) > 0:  # any dependency
                    if x == flags.AV_DOT_DIAMETER:
                        self.adapt_diameter.setEnabled(False)
                        self.adapt_diameter.setChecked(False)
                    elif x == flags.AV_PERIMETER:
                        self.adapt_av_perimeter.setEnabled(False)
                        self.adapt_av_perimeter.setChecked(False)
                    elif x == flags.AV_SURFACE_AREA:
                        self.adapt_av_area.setEnabled(False)
                        self.adapt_av_area.setChecked(False)
                    elif x == flags.TOTAL_SURFACE_AREA:
                        self.adapt_area.setEnabled(False)
                        self.adapt_area.setChecked(False)
                    elif x == flags.TOTAL_PERIMETER:
                        self.adapt_total_perimeter.setEnabled(False)
                        self.adapt_total_perimeter.setChecked(False)
                    elif x == flags.FIELD_AREA:
                        self.adapt_convex_hull.setEnabled(False)
                        self.adapt_convex_hull.setChecked(False)
                    elif x == flags.COVERAGE:
                        self.adapt_coverage.setEnabled(False)
                        self.adapt_coverage.setChecked(False)
                    elif x == flags.SPARSITY:
                        self.adapt_sparsity.setEnabled(False)
                        self.adapt_sparsity.setChecked(False)
                    elif x == flags.LOG_SIZE:
                        self.adapt_size.setEnabled(False)
                        self.adapt_size.setChecked(False)
                    elif x == flags.LOG_SPACING:
                        self.adapt_spacing.setEnabled(False)
                        self.adapt_spacing.setChecked(False)

        self.adapt_methods = selected

    @staticmethod
    def get_response(parent):

        dialog = SequenceDialog(parent)
        result = dialog.exec_()
        if result == QDialog.Accepted:
            SequenceDialog.extra_space = dialog.adapt_extra_space.value
            SequenceDialog.spacing_precision = dialog.adapt_spacing_presision.value
            SequenceDialog.adapt_FA2TA_ratio = dialog.adapt_fa2ta.value
            SequenceDialog.sequence_range = [dialog.adapt_range.value1, dialog.adapt_range.value2]
            return (dialog.adapt_methods,
                    SequenceDialog.sequence_range,
                    SequenceDialog.extra_space)
        else:
            return (None, None, None)



from . import fit as _fit
from .. import _lib

from .fit import change_fit_settings # make available


def average_diameter(dot_array, factor):
    if not isinstance(dot_array, _lib.DotArray):
        raise TypeError("Scaling diameter is not possible for {}.".format(
            type(dot_array).__name__))
    if factor == 1:
        return dot_array
    value = dot_array.properties.average_dot_diameter * factor
    return _fit.average_diameter(dot_array, value)

def average_rectangle_size(rect_array, factor):
    if not isinstance(rect_array, _lib.RectangleArray):
        raise TypeError("Scaling rectangle size is not possible for {}.".format(
            type(rect_array).__name__))
    if factor == 1:
        return rect_array
    value = rect_array.properties.average_rectangle_size * factor
    return _fit.average_rectangle_size(rect_array, value)


def total_surface_area(object_array, factor):
    if factor == 1:
        return object_array
    _lib._check_object_array(object_array)
    value = object_array.properties.total_surface_area * factor
    return _fit.total_surface_area(object_array, value)


def field_area(object_array, factor, precision=None):
    if factor == 1:
        return object_array
    _lib._check_object_array(object_array)
    value = object_array.properties.field_area * factor
    return _fit.field_area(object_array, value, precision=precision)

def coverage(object_array, factor,
             precision=None,
             FA2TA_ratio=None):
    if factor == 1:
        return object_array
    _lib._check_object_array(object_array)
    value = object_array.properties.converage * factor
    return _fit.coverage(object_array, value,
                        precision=precision,
                        FA2TA_ratio=FA2TA_ratio)

def average_perimeter(object_array, factor):
    if factor == 1:
        return object_array
    _lib._check_object_array(object_array)
    value = object_array.properties.average_perimeter * factor
    return _fit.average_perimeter(object_array, value)

def total_perimeter(object_array, factor):
    if factor == 1:
        return object_array
    _lib._check_object_array(object_array)
    value = object_array.properties.total_perimeter * factor
    return _fit.total_perimeter(object_array, value)

def average_surface_area(object_array, factor):
    if factor == 1:
        return object_array
    _lib._check_object_array(object_array)
    value = object_array.properties.average_surface_area * factor
    return _fit.average_surface_area(object_array, value)

def log_spacing(object_array, factor, precision=None):
    if factor == 1:
        return object_array
    _lib._check_object_array(object_array)
    value = object_array.properties.log_spacing * factor
    return _fit.log_spacing(object_array, value, precision=precision)

def log_size(object_array, factor):
    if factor == 1:
        return object_array
    _lib._check_object_array(object_array)
    value = object_array.properties.log_size * factor
    return _fit.log_size(object_array, value)

def sparcity(object_array, factor, precision=None):
    if factor == 1:
        return object_array
    _lib._check_object_array(object_array)
    value = object_array.properties.sparsity * factor
    return _fit.sparcity(object_array, value, precision=precision)

def visual_property(object_array, feature, factor):
    if factor == 1:
        return object_array
    _lib._check_object_array(object_array)
    value = object_array.properties.get(feature) * factor
    return _fit.visual_property(object_array, property_flag=feature, value=value)



"""
"""
try:
    from PyQt5.QtWidgets import QApplication as _QApplication
except:
    print("""ERROR: Running the PyNSN GUI requires the installation `PyQt5`.
Install `PyQt5` or  reinstall PyNSN via: `pip install pynsn[gui] -U`""")
    exit()

import sys as _sys
from ._gui.gui_main_window import GUIMainWindow as _GUIMainWindow


def start():
    app = _QApplication(_sys.argv)
    ex = _GUIMainWindow()
    ex.show()
    _sys.exit(app.exec_())




__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

from copy import deepcopy
from abc import ABCMeta, abstractmethod
from hashlib import md5
import json
import numpy as np
from scipy import spatial

from . import misc
from . import geometry
from . import shapes
from . import rng
from .array_tools import BrownianMotion
from ..visual_properties._properties import ArrayProperties
from ..exceptions import NoSolutionError
from ..visual_properties import fit
from .. import constants


class ArrayParameter(object):

    def __init__(self, target_area_radius,
                 min_dist_between=None,
                 min_dist_area_boarder=None):
        """Numpy Position lists with attributes for optimized for numpy calculations

        Abstract class for implementation of dot and rect
        """
        self.target_area_radius = target_area_radius
        if min_dist_between is None:
            self.min_dist_between = constants.DEFAULT_MIN_DIST_BETWEEN
        else:
            self.min_dist_between = min_dist_between
        if min_dist_area_boarder is None:
            self.min_dist_area_boarder = constants.DEFAULT_MIN_DIST_AREA_BOARDER
        else:
            self.min_dist_area_boarder = min_dist_area_boarder

    def as_dict(self):
        return {"type": type(self).__name__,
                "target_area_radius": self.target_area_radius,
                "min_dist_between": self.min_dist_between,
                "min_dist_area_boarder": self.min_dist_area_boarder}


class AttributeArray(ArrayParameter):
    """Class for attributes on two dimensional space"""

    def __init__(self, target_area_radius,
                 min_dist_between=None,
                 min_dist_area_boarder=None,
                 xy=None,
                 attributes=None):
        """Numpy Position lists with attributes for optimized for numpy calculations

        Abstract class for implementation of dot and rect
        """
        super().__init__(target_area_radius=target_area_radius,
                         min_dist_between=min_dist_between,
                         min_dist_area_boarder=min_dist_area_boarder)

        self._xy = np.array([])
        self._attributes = np.array([])
        self._properties = ArrayProperties(self)

        if xy is not None:
            self._append_xy_attribute(xy=xy, attributes=attributes)

    def _append_xy_attribute(self, xy, attributes=None):
        """returns number of added rows"""
        xy = misc.numpy_array_2d(xy)
        if not isinstance(attributes, (tuple, list)):
            attributes = [attributes] * xy.shape[0]

        if len(attributes) != xy.shape[0]:
            raise ValueError("Bad shaped data: attributes have not " +
                             "the same length as the coordinates")

        self._attributes = np.append(self._attributes, attributes)
        if len(self._xy) == 0:
            empty = np.array([]).reshape((0, 2))  # ensure good shape of self.xy
            self._xy = np.append(empty, xy, axis=0)
        else:
            self._xy = np.append(self._xy, xy, axis=0)
        self._properties.reset()
        return xy.shape[0]

    def __str__(self):
        prop_text = self._properties.as_text(extended_format=True)
        rtn = "- {}".format(type(self).__name__)
        rtn += "\n " + prop_text[1:]  # replace "-" with " "
        return rtn

    @property
    def xy(self):
        return self._xy

    @property
    def xy_rounded_integer(self):
        """rounded to integer"""
        return np.round(self._xy)

    @property
    def attributes(self):
        return self._attributes

    @property
    def properties(self):
        return self._properties

    @property
    def surface_areas(self):
        """per definition always zero"""
        return np.array([0] * len(self._xy))

    @property
    def perimeter(self):
        """per definition always zero"""
        return np.array([0] * len(self._xy))

    def set_attributes(self, attributes):
        """Set all attributes

        Parameter
        ---------
        attributes:  attribute (string) or list of attributes

        """

        if isinstance(attributes, (list, tuple)):
            if len(attributes) != self._properties.numerosity:
                raise ValueError("Length of attribute list does not adapt the " + \
                                 "size of the dot array.")
            self._attributes = np.array(attributes)
        else:
            self._attributes = np.array([attributes] * self._properties.numerosity)

    @property
    def hash(self):
        """md5_hash of positions and perimeter"""
        m = md5()
        m.update(
            self._xy.tobytes())  # to_byte required: https://stackoverflow.com/questions/16589791/most-efficient-property-to-hash-for-numpy-array
        try:
            m.update(self.perimeter.tobytes())
        except AttributeError:
            pass
        m.update(self._attributes.tobytes())
        return m.hexdigest()

    def get_center_of_field_area(self):
        """Center of all object positions
        """
        return geometry.center_of_positions(self.properties.convex_hull.xy)

    def clear(self):
        self._xy = np.array([])
        self._attributes = np.array([])
        self._properties.reset()

    def delete(self, index):
        self._xy = np.delete(self._xy, index, axis=0)
        self._attributes = np.delete(self._attributes, index)
        self._properties.reset()

    def copy(self, indices=None, deepcopy=True):
        """returns a (deep) copy of the dot array.

        It allows to copy a subset of dot only.

        """

        if len(self._xy) == 0:
            return AttributeArray(target_area_radius=self.target_area_radius,
                                  min_dist_between=self.min_dist_between,
                                  min_dist_area_boarder=self.min_dist_area_boarder)

        if indices is None:
            indices = list(range(len(self._xy)))

        if deepcopy:
            return AttributeArray(target_area_radius=self.target_area_radius,
                                  min_dist_between=self.min_dist_between,
                                  min_dist_area_boarder=self.min_dist_area_boarder,
                                  xy=self._xy[indices, :].copy(),
                                  attributes=self._attributes[indices].copy())
        else:
            return AttributeArray(target_area_radius=self.target_area_radius,
                                  min_dist_between=self.min_dist_between,
                                  min_dist_area_boarder=self.min_dist_area_boarder,
                                  xy=self._xy[indices, :],
                                  attributes=self._attributes[indices])

    def as_dict(self):
        """
        """
        d = super().as_dict()
        d.update({"xy": self._xy.tolist()})
        if len(self._attributes) > 0 and misc.is_all_equal(self._attributes):
            d.update({"attributes": self._attributes[0]})
        else:
            d.update({"attributes": self._attributes.tolist()})
        return d

    def read_from_dict(self, dict):
        """read dot array from dict"""
        self.target_area_radius = dict["target_area_radius"]
        self.min_dist_between = dict["min_dist_between"]
        self.min_dist_area_boarder = dict["min_dist_area_boarder"]
        self._xy = np.array(dict["xy"])
        if not isinstance(dict["attributes"], (list, tuple)):
            att = [dict["attributes"]] * self._properties.numerosity
        else:
            att = dict["attributes"]
        self._attributes = np.array(att)
        self._properties.reset()

    def json(self, indent=None, include_hash=False):
        """"""
        # override and extend as_dict not this function

        d = self.as_dict()
        if include_hash:
            d.update({"hash": self.hash})
        if not indent:
            indent = None
        return json.dumps(d, indent=indent)

    def save(self, json_file_name, indent=None, include_hash=False):
        """"""
        with open(json_file_name, 'w') as fl:
            fl.write(self.json(indent=indent, include_hash=include_hash))

    def load(self, json_file_name):
        # override and extend read_from_dict not this function
        with open(json_file_name, 'r') as fl:
            dict = json.load(fl)
        self.read_from_dict(dict)


class ABCObjectArray(AttributeArray, metaclass=ABCMeta):

    @property
    @abstractmethod
    def surface_areas(self):
        pass

    @property
    @abstractmethod
    def perimeter(self):
        pass

    @abstractmethod
    def as_dict(self):
        return super().as_dict()

    @abstractmethod
    def read_from_dict(self, dict):
        return super().read_from_dict()

    @abstractmethod
    def copy(self, indices=None, deepcopy=True):
        pass

    @abstractmethod
    def iter_objects(self, indices=None):
        pass

    def get_objects(self, indices=None):
        return list(self.iter_objects(indices=indices))

    @abstractmethod
    def add(self, something):
        pass

    @abstractmethod
    def find_objects(self, attribute):
        pass

    @abstractmethod
    def get_distances(self, ref_object):
        # override this method
        pass

    @abstractmethod
    def csv(self):
        pass

    @abstractmethod
    def mod_round_values(self):
        pass

    def join(self, object_array):
        """add another object arrays"""
        self.add(object_array.iter_objects())

    def get_distances_matrix(self, between_positions=False):
        """between position ignores the dot size"""
        if between_positions:
            return spatial.distance.cdist(self._xy, self._xy)
        # matrix with all distance between all points
        dist = np.asarray([self.get_distances(d) for d in self.iter_objects()])
        return dist

    def get_overlaps(self):
        """return pairs of indices of overlapping of objects and an array of the
        amount of overlap
        takes into account min_dist_between property

        """
        dist = misc.triu_nan(self.get_distances_matrix(between_positions=False),
                             k=1)
        overlap = np.where(dist < self.min_dist_between)
        return np.asarray(overlap).T, np.abs(dist[overlap])

    def get_center_of_mass(self):
        weighted_sum = np.sum(self._xy * self.perimeter[:, np.newaxis], axis=0)
        return weighted_sum / np.sum(self.perimeter)

    def mod_center_array_mass(self):
        self._xy = self._xy - self.get_center_of_mass()
        self._properties.reset()

    def mod_center_field_area(self):
        cxy = geometry.center_of_positions(self.properties.convex_hull.xy)
        self._xy = self._xy - cxy
        self._properties.reset()

    def get_free_position(self, ref_object,
                          in_neighborhood=False,
                          allow_overlapping=False,
                          inside_convex_hull=False,
                          occupied_space=None):
        """returns the copy of object of at a random free position

        raise exception if not found
        occupied space: see generator generate
        """

        if isinstance(ref_object, shapes.Dot):
            object_size = ref_object.diameter / 2.0
        elif isinstance(ref_object, shapes.Rectangle):
            object_size = max(ref_object.size)
        else:
            raise NotImplementedError("Not implemented for {}".format(
                type(ref_object).__name__))
        if occupied_space is not None and \
                not isinstance(occupied_space, ABCObjectArray):  # FIXME check
            raise TypeError("Occupied_space has to be a Dot or Rectangle Array or None.")

        area_rad = self.target_area_radius - self.min_dist_area_boarder - object_size
        rtn_object = deepcopy(ref_object)  # tested deepcopy required

        if in_neighborhood:
            random_walk = BrownianMotion(ref_object.xy, delta=2)
        else:
            random_walk = None

        cnt = 0
        while True:
            if cnt > constants.MAX_ITERATIONS:
                raise NoSolutionError(u"Can't find a free position")
            cnt += 1

            if in_neighborhood:
                rtn_object.xy = random_walk.next()
            else:
                rtn_object.xy = rng.generator.random(size=2) * 2 * area_rad - area_rad

            # is outside area
            if isinstance(ref_object, shapes.Dot):
                is_outside = area_rad <= rtn_object.polar_radius
            else:
                # Rect: check if one edge is outside
                is_outside = False
                for e in rtn_object.iter_edges():
                    if e.polar_radius >= area_rad:
                        is_outside = True
                        break

            if not is_outside and inside_convex_hull:
                # use only those that do not change the convex hull
                tmp_array = self.copy(deepcopy=True)
                tmp_array.add([rtn_object])
                is_outside = tmp_array.properties.convex_hull != \
                               self.properties.convex_hull
            if is_outside:
                if in_neighborhood:
                    random_walk.step_back()
                continue

            if not allow_overlapping:
                # find overlap
                dist = self.get_distances(rtn_object)
                if isinstance(occupied_space, ABCObjectArray):
                    dist = np.append(dist, occupied_space.get_distances(rtn_object))
                if sum(dist < self.min_dist_between) > 0:  # at least one is overlapping
                    continue
            return rtn_object

    def mod_shuffle_positions(self, allow_overlapping=False):
        """might raise an exception"""
        # find new position for each dot
        # mixes always all position (ignores dot limitation)

        all_objects = list(self.iter_objects())
        self.clear()
        for obj in all_objects:
            try:
                new = self.get_free_position(obj, in_neighborhood=False,
                                             allow_overlapping=allow_overlapping)
            except NoSolutionError as e:
                raise NoSolutionError("Can't shuffle dot array. No free positions found.")
            self.add([new])

    def get_outlier(self):
        """returns indices of object that stand out and array with the size
        of outstanding
        """

        xy = self.properties.convex_hull.xy
        sizes_outlying = np.hypot(xy[:, 0], xy[:, 1]) - \
                        (self.target_area_radius - self.min_dist_area_boarder)
        idx = sizes_outlying > 0
        return self.properties.convex_hull.object_indices[idx], sizes_outlying[idx]

    def get_number_deviant(self, change_numerosity, preserve_field_area=False):
        """number deviant
        """
        object_array = self.copy()
        new_num = self.properties.numerosity + change_numerosity
        fit.numerosity(object_array, value=new_num,
                       keep_convex_hull=preserve_field_area)
        return object_array

    def mod_remove_overlaps(self, keep_convex_hull=False, strict=False):
        """
        Returns
        Parameters
        ----------
        keep_convex_hull
        strict

        Returns
        -------
        rtn: boolean
            True, if field area has not been changed (in case strict=False)

        Notes
        -----

        TODO describe different algorithm for keep and not keep CH
        """

        warning_info = "Can't keep field area constant."
        old_fa = self.properties.field_area

        if not keep_convex_hull:
            # touch convex hull objects
            ids = list(range(len(self._xy)))
            for x in self.properties.convex_hull.object_indices_unique:
                self.mod_move_object(x, 0, (0, 0), push_other=True)
                ids.remove(x)
            # touch remaining ids
            for x in ids:
                # touch each object and push other
                self.mod_move_object(x, 0, (0, 0), push_other=True)

        else:
            overlaps = self.get_overlaps()[0]
            ch_idx = self.properties.convex_hull.object_indices_unique

            while len(overlaps):
                # do not replace convexhull objects and try to
                # take that one not on convex hull or raise error/warning
                if overlaps[0, 0] not in ch_idx:
                    idx = overlaps[0, 0]
                elif overlaps[0, 1] not in ch_idx:
                    idx = overlaps[0, 1]
                elif not strict:
                    # warning
                    idx = overlaps[0, 0]
                else:
                    raise NoSolutionError(warning_info)

                obj = next(self.iter_objects(idx))
                self.delete(idx)

                # search new pos: fist inside convex hull later outside
                found = None
                for inside_convex_hull in (True, False):
                    if strict and not inside_convex_hull:
                        continue
                    for in_neighborhood in (True, False):
                        if found is None:
                            try:
                                found = self.get_free_position(ref_object=obj,
                                                    in_neighborhood=in_neighborhood,
                                                    inside_convex_hull=inside_convex_hull)
                            except NoSolutionError as e:
                                found = None

                if found is None:
                    self.add([obj])
                    raise NoSolutionError("Can't find a solution for remove overlap")
                else:
                    self.add([found])
                    overlaps = self.get_overlaps()[0]

            self.properties.reset()

        # check convex hull change
        new_ch = self.properties.field_area
        if keep_convex_hull and old_fa != new_ch:
            if strict:
                raise NoSolutionError(warning_info)
            else:
                print("Warning: " + warning_info)

        return old_fa == new_ch

    def mod_move_object(self, object_id, distance, direction,
                        push_other=False):
        """

        Parameters
        ----------
        object_id
        distance
        direction: numeric (polar) or tuple, list or Point (cartesian)
            angle (numeric, polar angle coordinate) or a cartesian 2D coordinates indicating
            the direction towards the object should be moved

        Returns
        -------

        """

        try:
            ang = float(direction)
        except (TypeError, ValueError):
            try:
                ang = shapes.Point(x=direction[0], y=direction[1])
            except IndexError:
                ang = None
        if ang is None:
            raise TypeError("Direction has to be float or a 2D coordinate, that is, a "
                            "shapes.Point or  a tuple/list of two elements.")

        obj = next(self.iter_objects(indices=object_id))

        movement = shapes.Point()
        if isinstance(ang, float):
            movement.polar = (distance, ang)
        else:
            # "ang" isnot an angle it a point
            movement = ang - obj
            movement.polar_radius = distance

        self._xy[object_id, :] = self._xy[object_id, :] + movement.xy

        if push_other:
            # push overlapping object
            obj = next(self.iter_objects(indices=object_id))
            dist = self.get_distances(obj)
            for other_id in np.flatnonzero(dist<0):
                if other_id != object_id:
                    movement.xy = self._xy[other_id, :] - self._xy[object_id, :]
                    self.mod_move_object(other_id,
                                direction = movement.polar_angle,
                                distance = abs(dist[other_id]) + self.min_dist_between,
                                push_other=True)

        self.properties.reset()

    def mod_squeeze_to_area(self, push_other=True):
        """squeeze in target area to remove all standouts"""
        cnt = 0
        while True:
            cnt += 1
            if cnt > constants.MAX_ITERATIONS:
                raise NoSolutionError("Can't find a solution for squeezing")

            idx, size = self.get_outlier()
            if len(idx) == 0:
                return
            for object_id, size in zip(idx, size):
                self.mod_move_object(object_id, distance=size,
                                     direction = (0,0),
                                     push_other=push_other)

    def get_split_arrays(self):
        """returns a list of arrays
        each array contains all dots of with particular colour"""
        att = self._attributes
        att[np.where(att == None)] = "None"  # TODO check "is none"

        rtn = []
        for c in np.unique(att):
            if c is not None:
                da = self.copy(indices=0, deepcopy=False)  # fast. shallow copy with just one object
                da.clear()
                da.add(self.find_objects(attribute=c))
                rtn.append(da)
        return rtn

    def mod_realign(self, keep_convex_hull=False, strict=True):
        """

        Parameters
        ----------
        keep_convex_hull
        strict

        Returns
        -------
        field_area_unchanged, no_outlier

        """
        convex_hull_unchanged = self.mod_remove_overlaps(keep_convex_hull=keep_convex_hull,
                                                strict=strict)
        self.mod_center_field_area()

        has_no_outlier = len(self.get_outlier()[0]) == 0
        if not has_no_outlier:
            if keep_convex_hull:
                warning_info = "Can't keep field area constant."
                if strict:
                    raise NoSolutionError(warning_info)
                else:
                    print("Warning: " + warning_info)

            self.mod_squeeze_to_area(push_other=True)
            has_no_outlier = True
            convex_hull_unchanged = False

        return convex_hull_unchanged, has_no_outlier

# TODO  everywhere: file header doc and author information



from copy import copy

from .._lib import shapes
from ..distributions import PyNSNDistribution, _round_samples, Levels


class _Constant(object):

    def __init__(self, value):
        """Helper class to "sample" constance.

        Looks like a PyNSNDistribution, but sample returns just the constant

        Parameter:
        ----------
        constant : numeric
        """

        self.value = value

    def sample(self, n, round_to_decimals=None):
        return _round_samples([self.value] * n, round_to_decimals)

    def as_dict(self):
        return {"distribution": "Constant",
                "value": self.value}


def _make_distr(value):
    """helper
    returns a distribution or None, if None
    """

    if value is None:
        return None
    elif isinstance(value, PyNSNDistribution):
        return value
    elif isinstance(value, (list, tuple)):
        return Levels(levels=copy(value))
    else:
        return _Constant(value)


class AppearanceSampler(object):

    def __init__(self):

        self._distr_diameter = None
        self._distr_width = None
        self._distr_height = None
        self._distr_proportion = None
        self._distr_attributes = None

    @property
    def distr_diameter(self):
        return self._distr_diameter

    @property
    def distr_width(self):
        return self._distr_width

    @property
    def distr_height(self):
        return self._distr_height

    @property
    def distr_proportion(self):
        return self._distr_proportion

    @property
    def distr_attributes(self):
        return self._distr_attributes

    def set_appearance_dot(self, diameter, attributes=None):
        self._distr_width = None
        self._distr_height = None
        self._distr_proportion = None
        self._distr_diameter = _make_distr(diameter)
        self._distr_attributes = _make_distr(attributes)

    def set_appearance_rectangle(self, width=None, height=None,
                                 proportion=None, attributes=None):

        n_rect_parameter = sum([width is not None, height is not None,
                                proportion is not None])
        if n_rect_parameter == 1:
            raise TypeError("Define rectangle width and height or, alternatively, rectangle proportion together with "
                            "either width or height.")
        self._distr_diameter = None
        self._distr_width = _make_distr(width)
        self._distr_height = _make_distr(height)
        self._distr_proportion = _make_distr(proportion)
        self._distr_attributes = _make_distr(attributes)

    def is_appearance_set(self):
        return self._distr_diameter is not None or self._distr_width is not None or \
               self._distr_height is not None

    def sample(self, n, round_to_decimals=None):
        """return list objects (Dot or Rect) with random size
        all positions = (0,0)
        """
        if self._distr_attributes is not None:
            attributes = self._distr_attributes.sample(n)
        else:
            attributes = [None] * n
        if self._distr_diameter is not None:
            diameter = self._distr_diameter.sample(n)

            return [shapes.Dot(xy=(0, 0), diameter=dia, attribute=attr) \
                    for dia, attr in zip(diameter, attributes)]
        else:
            # Rect
            try:
                width = self._distr_width.sample(n)
            except AttributeError:
                width = None
            try:
                height = self._distr_height.sample(n)
            except AttributeError:
                height = None
            try:
                proportion = self._distr_proportion.sample(n)
            except AttributeError:
                proportion = None

            if height is None:
                height = width * proportion
            elif width is None:
                width = height / proportion

            if round_to_decimals is not None:
                width = _round_samples(width, round_to_decimals=round_to_decimals)
                height = _round_samples(width, round_to_decimals=round_to_decimals)

            return [shapes.Rectangle(xy=(0, 0), size=(w,h), attribute=attr)\
                    for w, h, attr in zip(width, height, attributes)]

    def as_dict(self):
        rtn = {}
        try:
            rtn.update({"diameter": self._distr_diameter.as_dict()})
        except AttributeError:
            pass
        try:
            rtn.update({"width": self._distr_width.as_dict()})
        except AttributeError:
            pass
        try:
            rtn.update({"height": self._distr_height.as_dict()})
        except AttributeError:
            pass
        try:
            rtn.update({"proportion": self._distr_proportion.as_dict()})
        except AttributeError:
            pass
        try:
            rtn.update({"attributes": self._distr_attributes.as_dict()})
        except AttributeError:
            pass
        return rtn



from os import path

class PictureFile(object):
    PICTURE_PREFIX = "p:"

    def __init__(self, filename):
        self._attribute = "{}{}".format(PictureFile.PICTURE_PREFIX, filename)

    @property
    def filename(self):
        return PictureFile.check_attribute(self._attribute)

    @property
    def attribute(self):
        return self._attribute

    @classmethod
    def check_attribute(cls, txt):
        """returns filename if `txt` (str) represent a picture attribute
        else returns None
        """
        if isinstance(txt, str) and \
                txt.startswith(PictureFile.PICTURE_PREFIX):
            return txt[len(PictureFile.PICTURE_PREFIX):]
        else:
            return None

    def check_file_exists(self):
        return path.isfile(self.filename)



import numpy as np
from scipy import spatial
from .._lib.geometry import cartesian2polar, polar2cartesian
from .. import _lib


class ConvexHullBaseClass(object):
    """convenient wrapper class for calculation of convex hulls
    """

    def __init__(self, xy):
        try:
            self._convex_hull = spatial.ConvexHull(xy)
        except spatial.QhullError:
            self._convex_hull = None

    @property
    def xy(self):
        if self._convex_hull is None:
            return np.array([])
        else:
            return self._convex_hull.points[self._convex_hull.vertices, :]

    @property
    def field_area(self):
        if self._convex_hull is None:
            return np.nan
        else:
            return self._convex_hull.volume

    @property
    def object_indices(self):
        return self._convex_hull.vertices

    def __eq__(self, other):
        """convex hulls are assumed to be identical if same field area is
        identical and convex comprises the same amount of points"""
        return len(self.xy) == len(other.xy) and \
                self.field_area == other.field_area


class ConvexHullPositions(ConvexHullBaseClass):
    """convenient wrapper class for calculation of convex hulls

    using just the positions (ignores the object size"
    """

    def __init__(self, object_array):
        assert isinstance(object_array, (_lib.AttributeArray, _lib.DotArray,
                                         _lib.RectangleArray))
        ConvexHullBaseClass.__init__(self, xy=object_array.xy)


class ConvexHull(ConvexHullBaseClass):
    """convenient wrapper class for calculation of convex hulls

    using outer positions, able to handle rects
    """

    def __init__(self, object_array):
        assert isinstance(object_array, (_lib.AttributeArray, _lib.DotArray,
                                         _lib.RectangleArray))

        if isinstance(object_array, _lib.DotArray):
            # centered polar coordinates
            minmax = np.array((np.min(object_array.xy, axis=0),
                               np.max(object_array.xy, axis=0)))
            center = np.reshape(minmax[1, :] - np.diff(minmax, axis=0) / 2, 2)  # center outer positions
            xy_centered = object_array.xy - center
            # outer positions
            polar_centered = cartesian2polar(xy_centered)
            polar_centered[:, 0] = polar_centered[:, 0] + (object_array.diameter / 2)
            outer_xy = polar2cartesian(polar_centered) + center

        elif isinstance(object_array, _lib.RectangleArray):
            # simple solution: get all edges
            edges = []
            for rect in object_array.iter_objects():
                edges.extend([e.xy for e in rect.iter_edges()])
            outer_xy = np.array(edges)

        else:  # generic attribute array
            outer_xy = object_array.xy

        ConvexHullBaseClass.__init__(self, xy=outer_xy)
        self._is_rect_array = isinstance(object_array, _lib.RectangleArray)

    @property
    def object_indices(self):
        if self._is_rect_array:
            return np.int16(self._convex_hull.vertices/4)
        else:
            return self._convex_hull.vertices

    @property
    def object_indices_unique(self):
        return np.unique(self.object_indices)



"""
Rectangle Array
"""

__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import numpy as np

from .base_classes import ABCObjectArray
from .._lib import misc
from .._lib.shapes import Rectangle, Point


class RectangleArray(ABCObjectArray):
    """
    """

    def __init__(self,
                 target_area_radius,
                 min_dist_between=None,
                 min_dist_area_boarder=None,
                 xy=None,
                 sizes=None,
                 attributes=None):
        """Rectangular array is restricted to a certain area, it has a target area
        and a minimum gap.

        This properties allows shuffling free position and adapting
        properties.

        """
        super().__init__(xy=xy, attributes=attributes,
                         target_area_radius=target_area_radius,
                         min_dist_between=min_dist_between,
                         min_dist_area_boarder=min_dist_area_boarder)
        self._sizes = np.array([])
        if sizes is not None:
            self._append_sizes(sizes)

        if self._xy.shape[0] != self._sizes.shape[0]:
            raise ValueError("Bad shaped data: " +
                             u"xy has not the same length as sizes array")

    def _append_sizes(self, sizes):
        """returns number of added rows"""
        sizes = misc.numpy_array_2d(sizes)
        if len(self._sizes) == 0:
            empty = np.array([]).reshape((0, 2))  # ensure good shape of self.xy
            self._sizes = np.append(empty, sizes, axis=0)
        else:
            self._sizes = np.append(self._sizes, sizes, axis=0)
        return sizes.shape[0]

    def add(self, rectangles):
        """append one dot or list of dots"""
        try:
            rectangles = list(rectangles)
        except TypeError:
            rectangles = [rectangles]

        for r in rectangles:
            assert isinstance(r, Rectangle)
            self._append_xy_attribute(xy=r.xy, attributes=r.attribute)
            self._append_sizes((r.width, r.height))
        self.properties.reset()

    @property
    def sizes(self):
        return self._sizes

    @property
    def surface_areas(self):
        # a = w*h
        return self._sizes[:, 0] * self._sizes[:, 1]

    @property
    def perimeter(self):
        return 2 * (self._sizes[:, 0] + self._sizes[:, 1])

    def mod_round_values(self, decimals=0, int_type=np.int32):
        """Round values of the array."""

        if decimals is None:
            return
        self._xy = misc.numpy_round2(self._xy, decimals=decimals,
                                     int_type=int_type)
        self._sizes = misc.numpy_round2(self._sizes, decimals=decimals,
                                        int_type=int_type)

    def as_dict(self):
        """
        """
        d = super().as_dict()
        d.update({"sizes": self._sizes.tolist()})
        return d

    def read_from_dict(self, the_dict):
        """read rectangle array from dict"""
        super().read_from_dict(the_dict)
        self._sizes = np.array(the_dict["sizes"])

    def clear(self):
        super().clear()
        self._sizes = np.array([])

    def delete(self, index):
        super().delete(index)
        self._sizes = np.delete(self._sizes, index, axis=0)

    def copy(self, indices=None, deepcopy=True):
        """returns a (deep) copy of the dot array.

        It allows to copy a subset of dot only.

        """

        if len(self._xy) == 0:
            return RectangleArray(
                target_area_radius=self.target_area_radius,
                min_dist_area_boarder=self.min_dist_area_boarder,
                min_dist_between=self.min_dist_between)
        if indices is None:
            indices = list(range(len(self._xy)))

        if deepcopy:
            return RectangleArray(
                target_area_radius=self.target_area_radius,
                min_dist_between=self.min_dist_between,
                min_dist_area_boarder=self.min_dist_area_boarder,
                xy=self._xy[indices, :].copy(),
                sizes=self._sizes[indices].copy(),
                attributes=self._attributes[indices].copy())
        else:
            return RectangleArray(
                target_area_radius=self.target_area_radius,
                min_dist_between=self.min_dist_between,
                min_dist_area_boarder=self.min_dist_area_boarder,
                xy=self._xy[indices, :],
                sizes=self._sizes[indices],
                attributes=self._attributes[indices])

    def _xy_distances(self, rect):
        """return distances on both axes between rectangles and reference rec.
         negative number indicates overlap edges along that dimension.
        """
        if len(self._xy) == 0:
            return np.array([])
        else:
            pos_dist = np.abs(self._xy - rect.xy)
            max_not_overlap_dist = (self.sizes + rect.size) / 2
            dist = pos_dist - max_not_overlap_dist
            return dist # FIXME intebnsive test distance function rect (also get_distance)

    def get_distances(self, rect):
        """Euclidean Distances toward a single Rectangle
        negative numbers indicate overlap

        Returns
        -------
        distances : numpy array of distances
        """
        assert isinstance(rect, Rectangle)
        if len(self._xy) == 0:
            return np.array([])
        else:
            d_xy = self._xy_distances(rect)
            eucl_dist = np.hypot(d_xy[:, 0], d_xy[:, 1])
            for i, n_neg in enumerate(np.sum(d_xy < 0, axis=1)):
                if n_neg == 2:
                    # two dimensions overlap -> calc distance and make negative
                    eucl_dist[i] = -1 * eucl_dist[i]
                elif n_neg == 1:
                    # one dimension overlaps -> set to zero
                    if d_xy[i, 0] < 0:
                        x = 0
                    else:
                        x = d_xy[i, 0]
                    if d_xy[i, 1] < 0:
                        y = 0
                    else:
                        y = d_xy[i, 1]
                    eucl_dist[i] = np.hypot(x, y)

            return eucl_dist

    def iter_objects(self, indices=None):
        """iterate over all or a part of the objects

        Parameters
        ----------
        indices

        Notes
        -----
        To iterate all object you might all use the class iterator __iter__:
        >>> for obj in my_array:
        >>>    print(obj)
        """
        if indices is None:
            data = zip(self._xy, self._sizes, self._attributes)
        else:
            try:
                indices = list(indices)  # check if iterable
            except TypeError:
                indices = [indices]
            data = zip(self._xy[indices, :], self._sizes[indices],
                       self._attributes[indices])

        for xy, s, att in data:
            rtn = Rectangle(xy=xy, size=s, attribute=att)
            yield rtn

    def find_objects(self, size=None, attribute=None, edge=None):
        """returns indices of found objects

        2D-tuple
        """
        rtn = []
        for i in range(len(self._sizes)):
            if (size is not None and self._sizes[i] != size) or \
                    (attribute is not None and self._attributes[i] != attribute):
                continue
            rtn.append(i)

        if edge is not None:
            return rtn
        elif isinstance(edge, Point):
            new_rtn = []
            for i, rect in zip(rtn, self.iter_objects(indices=rtn)):
                if edge in list(rect.iter_edges()):
                    new_rtn.append(i)
            return new_rtn
        else:
            raise TypeError("edge has to be of type Points")

    def csv(self, variable_names=True, hash_column=True,
            attribute_column=False):
        """Return the rectangle array as csv text

        Parameter
        ---------
        variable_names : bool, optional
            if True variable name will be printed in the first line
        """
        size_dict = {"width": self._sizes[:, 0], "height": self._sizes[:, 1]}
        if attribute_column:
            attr = self.attributes
        else:
            attr = None
        if hash_column:
            array_hash = self.hash
        else:
            array_hash = None
        return misc.make_csv(xy=self._xy,
                             size_data_dict=size_dict,
                             attributes=attr, array_hash=array_hash,
                             make_variable_names=variable_names)

    def get_split_arrays(self):
        """returns a list of arrays
        each array contains all dots of with particular colour"""
        att = self._attributes
        att[np.where(att == None)] = "None"  # TODO check "is none"

        rtn = []
        for c in np.unique(att):
            if c is not None:
                da = RectangleArray(target_area_radius=self.target_area_radius,
                                    min_dist_between=self.min_dist_between,
                                    min_dist_area_boarder=self.min_dist_area_boarder)
                da.add(self.find_objects(attribute=c))
                rtn.append(da)
        return rtn


# FIXME overlapping does not work for rectangles
# FIXME picture random size distribution



"""
"""
from PyQt5.QtCore import Qt, QCoreApplication
from PyQt5.QtGui import QFont, QIntValidator, QDoubleValidator
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QLabel, QLineEdit, QProgressDialog


class LabeledInput(object):

    def __init__(self, label, text, width_label=180, width_edit=70, case_sensitive=True):

        self.label = QLabel(label)
        self.label.setFixedWidth(width_label)
        self.edit = QLineEdit()
        self.edit.setFixedWidth(width_edit)
        self.edit.setAlignment(Qt.AlignRight)
        self.case_sensitive = case_sensitive
        self.text = text

    @property
    def text(self):
        rtn = str(self.edit.text())
        if not self.case_sensitive:
            return rtn.lower()
        else:
            return rtn

    @text.setter
    def text(self, v):
        if not self.case_sensitive:
            self.edit.setText(str(v).lower())
        else:
            self.edit.setText(str(v))

    def layout(self, vertical=False):
        if vertical:
            layout = QVBoxLayout()
        else:
            layout = QHBoxLayout()
        layout.addWidget(self.label)
        layout.addWidget(self.edit)
        return layout

    def setVisible(self, bool):
        self.edit.setVisible(bool)
        self.label.setVisible(bool)


class LabeledNumberInput(LabeledInput):
    def __init__(self, label, value, width_label=180, width_edit=70,
                 integer_only=True, min=None, max=None):

        LabeledInput.__init__(self, label=label, text="", width_label=width_label,
                              width_edit=width_edit)
        self.integer_only = integer_only
        if integer_only:
            self.edit.setValidator(QIntValidator())
        else:
            self.edit.setValidator(QDoubleValidator())
        self.min = min
        self.max = max
        self.value = value

    @property
    def value(self):
        if self.integer_only:
            rtn = int(self.edit.text())
        else:
            rtn = float(self.edit.text())
        if self.min is not None and rtn < self.min:
            rtn = self.min
            self.value = rtn
        if self.max is not None and rtn > self.max:
            rtn = self.max
            self.value = rtn

        return rtn

    @value.setter
    def value(self, v):
        if self.min is not None and v < self.min:
            v = self.min
        if self.max is not None and v > self.max:
            v = self.max
        if self.integer_only:
            v = int(v)
        else:
            v = float(v)
        self.edit.setText(str(v))


class LabeledNumberInputTwoValues(object):
    def __init__(self, label, value1, value2, width_label=180, width_edits=35,
                 integer_only=True):
        self.input1 = LabeledNumberInput(label=label, value=value1, width_label=width_label, width_edit=width_edits,
                                         integer_only=integer_only)

        self.input2 = LabeledNumberInput(label="", value=value2, width_label=width_label, width_edit=width_edits,
                                         integer_only=integer_only)

    @property
    def value1(self):
        return self.input1.value

    @value1.setter
    def value1(self, v):
        self.input1.value = v

    @property
    def value2(self):
        return self.input2.value

    @value2.setter
    def value2(self, v):
        self.input2.value = v

    def layout(self, vertical=False):
        layout = self.input1.layout(vertical)
        layout.addWidget(self.input2.edit)
        return layout


def heading(text):
    boldFont = QFont()
    boldFont.setBold(True)
    heading = QLabel(text)
    heading.setFont(boldFont)
    heading.setStyleSheet("QLabel { color : black; }")
    return heading


class NumberInput(object):

    def __init__(self, value, width_edit=70, integer_only=False, min=None, max=None):

        self.edit = QLineEdit()
        self.edit.setFixedWidth(width_edit)
        self.edit.setAlignment(Qt.AlignRight)
        self.integer_only = integer_only
        if integer_only:
            self.edit.setValidator(QIntValidator())
        else:
            self.edit.setValidator(QDoubleValidator())
        self.min = min
        self.max = max
        self.value = value

    @property
    def value(self):
        if self.integer_only:
            rtn = int(self.edit.text())
        else:
            rtn = float(self.edit.text())
        if self.min is not None and rtn < self.min:
            rtn = self.min
            self.value = rtn
        if self.max is not None and rtn > self.max:
            rtn = self.max
            self.value = rtn

        return rtn

    @value.setter
    def value(self, v):
        if self.min is not None and v < self.min:
            v = self.min
        if self.max is not None and v > self.max:
            v = self.max
        if self.integer_only:
            v = int(v)
        else:
            v = float(v)
        self.edit.setText(str(v))


def progressbar_iterator(iteratable, n_elements, label, win_title):
    # iterator function with prgress bar
    dialog = QProgressDialog()
    dialog.setMaximum(100)
    dialog.setLabelText(label)
    dialog.setWindowTitle(win_title)
    dialog.open()
    for cnt, item in enumerate(iteratable):
        p = (cnt * 100) // n_elements
        QCoreApplication.instance().processEvents()
        if dialog.wasCanceled():
            raise StopIteration
        dialog.setValue(p)
        yield (item)

    dialog.close()



__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

from PIL import Image as _Image
from PIL import ImageDraw as _ImageDraw
import numpy as _np
from .._lib.geometry import cartesian2image_coordinates as _c2i_coord
from .. import _lib
from . import _array_draw


# FIXME pillow supports no alpha/opacity


def create(object_array, colours=None, antialiasing=True):
    # ImageParameter
    """use PIL colours (see PIL.ImageColor.colormap)

    returns pil image

    antialiasing: Ture or integer

    default_dot_colour: if colour is undefined in _lib
    """

    return _PILDraw().create_image(object_array=object_array,
                                   colours=colours,
                                   antialiasing=antialiasing)


class _PILDraw(_array_draw.ArrayDraw):

    @staticmethod
    def get_squared_image(image_width, background_colour, **kwargs):
        # filename not used for pil images
        return _Image.new("RGBA", (image_width, image_width), color=background_colour)

    @staticmethod
    def scale_image(image, scaling_factor):
        im_size = int(image.size[0] / scaling_factor)
        return image.resize((im_size, im_size), _Image.LANCZOS)

    @staticmethod
    def draw_shape(img, shape, opacity, scaling_factor):
        # FIXME opacity is ignored (not yet supported)
        # draw object
        shape.xy = _c2i_coord(_np.asarray(shape.xy) * scaling_factor, img.size[0]).tolist()
        attr = shape.get_attribute_object()

        if isinstance(shape, _lib.Dot):
            shape.diameter = shape.diameter * scaling_factor
            r = shape.diameter / 2
            _ImageDraw.Draw(img).ellipse((shape.x - r, shape.y - r,
                                          shape.x + r, shape.y + r),
                                         fill=attr.colour)
        elif isinstance(shape, _lib.Rectangle):
            tmp = _np.asarray(shape.size) * scaling_factor
            shape.size = tmp.tolist()
            if isinstance(attr, _lib.PictureFile):
                # picture
                shape_size = (round(shape.width), round(shape.height))
                target_box = (round(shape.left), round(shape.bottom),  # FIXME ceil?
                              round(shape.right), round(shape.top))  # reversed y axes
                pict = _Image.open(attr.filename, "r")
                if pict.size[0] != shape_size[0] or pict.size[1] != shape_size[1]:
                    pict = pict.resize(shape_size, resample=_Image.ANTIALIAS)

                tr_layer = _Image.new('RGBA', img.size, (0, 0, 0, 0))
                tr_layer.paste(pict, target_box)
                res = _Image.alpha_composite(img, tr_layer)
                img.paste(res)
            else:
                # rectangle shape
                _ImageDraw.Draw(img).rectangle((shape.left, shape.top,
                                                shape.right, shape.bottom),
                                               fill=attr.colour)  # FIXME decentral shapes seems to be bit larger than with pyplot

        else:
            raise NotImplementedError("Shape {} NOT YET IMPLEMENTED".format(type(shape)))

    @staticmethod
    def draw_convex_hull(img, points, convex_hull_colour,  opacity,
                         scaling_factor):
        # FIXME opacity is ignored (not yet supported)
        points = _c2i_coord(points * scaling_factor, img.size[0])
        last = None
        draw = _ImageDraw.Draw(img)
        for p in _np.append(points, [points[0]], axis=0):
            if last is not None:
                draw.line(_np.append(last, p).tolist(),
                          width=2,
                          fill=convex_hull_colour.colour)
            last = p



__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import pygame as _pygame

from . import _colour
from . import pil_image as _pil_image
from .. import _lib


def create(object_array,
           colours=None,
           antialiasing=True):
    _lib._check_object_array(object_array)
    if colours is None:
        colours = _colour.ImageColours()
    if not isinstance(colours, _colour.ImageColours):
        raise TypeError("Colours must be of type image.ImageColours")

    img = _pil_image.create(object_array=object_array,
                            colours=colours,
                            antialiasing=antialiasing)

    return _pygame.image.fromstring(img.tobytes(), img.size, img.mode)



"""
"""
from multiprocessing import Pool

from PIL.ImageQt import ImageQt
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QDialog, QHBoxLayout, QLabel, QSlider

from ..image import pil_image
from . import misc


def _map_make_image(x):
    da, colours, aa = x
    return pil_image.create(object_array=da, colours=colours,
                            antialiasing=aa)


class SequenceDisplay(QDialog):

    def __init__(self, parent, da_sequence, start_numerosity, image_colours,
                 antialiasing):
        super(SequenceDisplay, self).__init__(parent)

        self.setWindowTitle("Dot Array Sequence")
        self.da_sequence = da_sequence
        self.pixmap_width = da_sequence.dot_arrays[0].target_area_radius * 2

        self.picture_field = QLabel(self)
        self.picture_field.setFixedSize(self.pixmap_width, self.pixmap_width)

        self.slider = QSlider(Qt.Vertical)
        num_range = da_sequence.min_max_numerosity
        self.slider.setMinimum(num_range[0])
        self.slider.setMaximum(num_range[1])
        self.slider.setValue(start_numerosity)
        self.slider.valueChanged.connect(self.action_slider_change)

        hlayout = QHBoxLayout()
        hlayout.addWidget(self.picture_field)
        hlayout.addWidget(self.slider)
        self.setLayout(hlayout)

        # make images
        image_colours = [image_colours] * len(self.da_sequence.dot_arrays)
        antialiasing = [antialiasing] * len(self.da_sequence.dot_arrays)

        iter_images = Pool().imap(_map_make_image,
                                  zip(self.da_sequence.dot_arrays,
                                      image_colours, antialiasing))
        progbar_iter = misc.progressbar_iterator(iter_images,
                                                 n_elements=len(self.da_sequence.dot_arrays),
                                                 label="make images", win_title="Dot Array Sequence")

        self.pixmaps = list(map(lambda im: QPixmap.fromImage(ImageQt(im)), progbar_iter))
        self.updateUI()

    def updateUI(self):
        num = self.slider.value()
        idx = self.da_sequence.numerosity_idx[num]
        feat = self.da_sequence.dot_arrays[idx]._properties.as_text(
            extended_format=False, with_hash=False)
        self.setWindowTitle(feat)
        self.picture_field.setPixmap(self.pixmaps[idx])
        self.adjustSize()

    def action_slider_change(self):
        self.updateUI()



__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import math as _math
import pygame as _pygame
from multiprocessing import Pool as _Pool

from expyriment.misc import Clock as _Clock
from expyriment.stimuli import Canvas as _Canvas
from . import _colour
from . import pil_image as _pil_image
from .. import _lib


class ExprimentDotArray(_Canvas):

    def __init__(self, object_array,
                 colours=None,
                 position=(0, 0),
                 antialiasing=True):

        _lib._check_object_array(object_array)
        if colours is None:
            colours = _colour.ImageColours()
        if not isinstance(colours, _colour.ImageColours):
            raise TypeError("Colours must be of type image.ImageColours")

        _Canvas.__init__(self, size=(0, 0), position=position)
        self.dot_array = object_array
        self.colours = colours
        self.antialiasing = antialiasing
        self._image = None

    @property
    def image(self):
        if self._image is None:
            self._create_pil_image()

        return self._image

    def _create_pil_image(self):
        self._image = _pil_image.create(object_array=self.dot_array,
                                        colours= self.colours,
                                        antialiasing=self.antialiasing) #TODO gabor filter
        return self._image

    def _create_surface(self):
        self._size = self.image.size
        return _pygame.image.frombuffer(self.image.tobytes(),
                                       self.image.size,
                                       self.image.mode)


class ExpyrimentDASequence(object):

    def __init__(self, da_sequence,
                 # pil_image_generator TODO better using generator
                 position=(0, 0),
                 colours = _colour.ImageColours(),
                 antialiasing=None,
                 make_pil_images_now=False,
                 multiprocessing=False):

        if not isinstance(colours, _colour.ImageColours):
            raise TypeError("Colours must be a ImageColours instance")

        self.da_sequence = da_sequence
        self.stimuli = []
        self.position = position
        self.antialiasing = antialiasing

        for da in self.da_sequence.dot_arrays:
            stim = ExprimentDotArray(object_array=da, position=position,
                                     colours=colours,
                                     antialiasing=antialiasing)
            self.stimuli.append(stim)

        if make_pil_images_now:

            if not multiprocessing:
                list(map(lambda x: x._create_pil_image(), self.stimuli))
                self._make_image_process = None
            else:
                p = _Pool()

                for c, pil_im in enumerate(p.imap(ExpyrimentDASequence._make_stimuli_map_helper, self.stimuli)):
                    self.stimuli[c]._image = pil_im
                p.close()
                p.join()

    def get_stimulus_numerosity(self, number_of_dots):
        """returns image with a particular numerosity"""
        try:
            return self.stimuli[self.da_sequence.numerosity_idx[number_of_dots]]
        except:
            return None

    @property
    def is_preloaded(self):
        for x in reversed(self.stimuli):
            if not x.is_preloaded:
                return False
        return True

    def preload(self, until_percent=100, time=None, do_not_return_earlier=False):
        """
        preloaded all _lib stimuli

        Note: this will take a while!

        preload certain percent or or a time.

        """
        if until_percent > 0 and until_percent < 100:
            last = int(_math.floor(until_percent * len(self.stimuli) / 100.0))
        elif until_percent == 0:
            last = 0
        else:
            last = len(self.stimuli)
        cl = _Clock()

        try:
            for x in self.stimuli[:last]:
                if not x.is_preloaded and (time is None or cl.time < time):
                    x.preload()
            rtn = True
        except:
            rtn = False

        if do_not_return_earlier and time is not None:
            cl.wait(time - cl.time)
        return rtn

    def unload(self):
        """
        returns array of preloaded dot_array_sequence
        """

        try:
            list(map(lambda x: x.unload(), self.stimuli))
            return True
        except:
            return False

    @staticmethod
    def _make_stimuli_map_helper(x):
        return x._create_pil_image()



"""
The named colour are the 140 HTML colour names:
   see https://www.w3schools.com/colors/colors_names.asp
"""
from collections import OrderedDict
from functools import total_ordering
from .._lib import misc

_NUMERALS = '0123456789abcdefABCDEF'
_HEXDEC = {v: int(v, 16) for v in (x + y for x in _NUMERALS for y in _NUMERALS)}


@total_ordering
class Colour(object):

    def __init__(self, colour, default=None):
        """if colour is unknown or None, default colour is used

        Thus, `Colour(variable, default="red")` will result in red is variable
        is None or Unknown. If default is not defined a `None colour` will be
        returned.
        """
        if colour is None:
            colour = default

        self._colour = None
        try:
            self.set(colour)
        except TypeError:  # colour is not a valid colour
            self.set(default)

    def __repr__(self):
        return "Colour({})".format(self.colour)

    def __hash__(self):
        return hash(self._colour)

    def __lt__(self, other):
        return self._colour < other._colour

    def __eq__(self, other):
        return self._colour == other._colour

    def __ne__(self, other):
        return self._colour != other._colour

    @property
    def colour(self):
        return self._colour

    def set(self, value):
        if value is None:
            self._colour = None
        elif isinstance(value, Colour):
            self._colour = value._colour
        else:
            try:
                Colour.hextriplet2rgb(value)  # check if valid hextriplet
                if value[0] != "#":
                    value = "#" + value
                self._colour = value.upper()
            except:
                try:
                    self._colour = Colour.NAMED_COLOURS[value]
                except:
                    try:
                        self._colour = Colour.rgb2hextriplet(value)
                    except:
                        raise TypeError("Incorrect colour " + \
                                        "('{}')!\n Use RGB tuple, hex triplet or a colour name from Colour.NAMED_COLOURS.".format(
                                            value))

    @property
    def rgb(self):
        return Colour.hextriplet2rgb(self._colour)

    def rgb_alpha(self, alpha):
        """if alpha can be float <= 1.0 or an integer between [0, 255]"""

        if isinstance(alpha, float):
            if alpha < 0. or alpha > 1.:
                raise TypeError("If alpha is a float, it has to be between 0 and 1.")
            alpha = round(alpha * 255)
        elif isinstance(alpha, int):
            if alpha < 0 or alpha > 255:
                raise TypeError("If alpha is an int, it has to be between 0 and 255.")
        else:
            raise TypeError("If alpha has to be a float or int and not {}.".format(
                type(alpha)))

        a = Colour.hextriplet2rgb(self._colour)
        return Colour.hextriplet2rgb(self._colour) + (alpha,)

    @staticmethod
    def hextriplet2rgb(hextriplet):
        ht = hextriplet.lstrip("#")
        return _HEXDEC[ht[0:2]], _HEXDEC[ht[2:4]], _HEXDEC[ht[4:6]]

    @staticmethod
    def rgb2hextriplet(rgb, uppercase=True):
        if len(rgb) != 3:
            raise TypeError("rgb must be a list of three values.")
        if uppercase:
            lettercase = 'X'
        else:
            lettercase = 'x'
        return "#" + format(rgb[0] << 16 | rgb[1] << 8 | rgb[2], '06' + lettercase)

    NAMED_COLOURS = {

        'aliceblue': '#F0F8FF',
        'antiquewhite': '#FAEBD7',
        'aqua': '#00FFFF',
        'aquamarine': '#7FFFD4',
        'azure': '#F0FFFF',
        'beige': '#F5F5DC',
        'bisque': '#FFE4C4',
        'black': '#000000',
        'blanchedalmond': '#FFEBCD',
        'blue': '#0000FF',
        'blueviolet': '#8A2BE2',
        'brown': '#A52A2A',
        'burlywood': '#DEB887',
        'cadetblue': '#5F9EA0',
        'chartreuse': '#7FFF00',
        'chocolate': '#D2691E',
        'coral': '#FF7F50',
        'cornflowerblue': '#6495ED',
        'cornsilk': '#FFF8DC',
        'crimson': '#DC143C',
        'cyan': '#00FFFF',
        'darkblue': '#00008B',
        'darkcyan': '#008B8B',
        'darkgoldenrod': '#B8860B',
        'darkgray': '#A9A9A9',
        'darkgreen': '#006400',
        'darkkhaki': '#BDB76B',
        'darkmagenta': '#8B008B',
        'darkolivegreen': '#556B2F',
        'darkorange': '#FF8C00',
        'darkorchid': '#9932CC',
        'darkred': '#8B0000',
        'darksalmon': '#E9967A',
        'darkseagreen': '#8FBC8F',
        'darkslateblue': '#483D8B',
        'darkslategray': '#2F4F4F',
        'darkturquoise': '#00CED1',
        'darkviolet': '#9400D3',
        'deeppink': '#FF1493',
        'deepskyblue': '#00BFFF',
        'dimgray': '#696969',
        'dodgerblue': '#1E90FF',
        'firebrick': '#B22222',
        'floralwhite': '#FFFAF0',
        'forestgreen': '#228B22',
        'fuchsia': '#FF00FF',
        'gainsboro': '#DCDCDC',
        'ghostwhite': '#F8F8FF',
        'gold': '#FFD700',
        'goldenrod': '#DAA520',
        'gray': '#808080',
        'green': '#008000',
        'greenyellow': '#ADFF2F',
        'honeydew': '#F0FFF0',
        'hotpink': '#FF69B4',
        'indianred': '#CD5C5C',
        'indigo': '#4B0082',
        'ivory': '#FFFFF0',
        'khaki': '#F0E68C',
        'lavender': '#E6E6FA',
        'lavenderblush': '#FFF0F5',
        'lawngreen': '#7CFC00',
        'lemonchiffon': '#FFFACD',
        'lightblue': '#ADD8E6',
        'lightcoral': '#F08080',
        'lightcyan': '#E0FFFF',
        'lightgoldenrodyellow': '#FAFAD2',
        'lightgreen': '#90EE90',
        'lightgray': '#D3D3D3',
        'lightpink': '#FFB6C1',
        'lightsalmon': '#FFA07A',
        'lightseagreen': '#20B2AA',
        'lightskyblue': '#87CEFA',
        'lightslategray': '#778899',
        'lightsteelblue': '#B0C4DE',
        'lightyellow': '#FFFFE0',
        'lime': '#00FF00',
        'limegreen': '#32CD32',
        'linen': '#FAF0E6',
        'magenta': '#FF00FF',
        'maroon': '#800000',
        'mediumaquamarine': '#66CDAA',
        'mediumblue': '#0000CD',
        'mediumorchid': '#BA55D3',
        'mediumpurple': '#9370DB',
        'mediumseagreen': '#3CB371',
        'mediumslateblue': '#7B68EE',
        'mediumspringgreen': '#00FA9A',
        'mediumturquoise': '#48D1CC',
        'mediumvioletred': '#C71585',
        'midnightblue': '#191970',
        'mintcream': '#F5FFFA',
        'mistyrose': '#FFE4E1',
        'moccasin': '#FFE4B5',
        'navajowhite': '#FFDEAD',
        'navy': '#000080',
        'oldlace': '#FDF5E6',
        'olive': '#808000',
        'olivedrab': '#6B8E23',
        'orange': '#FFA500',
        'orangered': '#FF4500',
        'orchid': '#DA70D6',
        'palegoldenrod': '#EEE8AA',
        'palegreen': '#98FB98',
        'paleturquoise': '#AFEEEE',
        'palevioletred': '#DB7093',
        'papayawhip': '#FFEFD5',
        'peachpuff': '#FFDAB9',
        'peru': '#CD853F',
        'pink': '#FFC0CB',
        'plum': '#DDA0DD',
        'powderblue': '#B0E0E6',
        'purple': '#800080',
        'red': '#FF0000',
        'rosybrown': '#BC8F8F',
        'royalblue': '#4169E1',
        'saddlebrown': '#8B4513',
        'salmon': '#FA8072',
        'sandybrown': '#FAA460',
        'seagreen': '#2E8B57',
        'seashell': '#FFF5EE',
        'sienna': '#A0522D',
        'silver': '#C0C0C0',
        'skyblue': '#87CEEB',
        'slateblue': '#6A5ACD',
        'slategray': '#708090',
        'snow': '#FFFAFA',
        'springgreen': '#00FF7F',
        'steelblue': '#4682B4',
        'tan': '#D2B48C',
        'teal': '#008080',
        'thistle': '#D8BFD8',
        'tomato': '#FF6347',
        'turquoise': '#40E0D0',
        'violet': '#EE82EE',
        'wheat': '#F5DEB3',
        'white': '#FFFFFF',
        'whitesmoke': '#F5F5F5',
        'yellow': '#FFFF00',
        'yellowgreen': '#9ACD32',
        'expyriment_orange': '#FF9632',
        'expyriment_purple': '#A046FA'
    }


class ImageColours(object):
    COL_TARGET_AREA = "#FFF0D9"
    COL_DEFAULT_OBJECT = "darkgreen"
    OPACITY_OBJECT = 1
    OPACITY_GUIDES = 0.5

    def __init__(self,
                 target_area=None,
                 field_area_positions=None,
                 field_area=None,
                 center_of_field_area=None,
                 center_of_mass=None,
                 background=None,
                 default_object_colour=None,
                 opacity_object=None,
                 opacity_guides=None
                 ):

        self.target_area = Colour(target_area,
                                  default=ImageColours.COL_TARGET_AREA)
        self.field_area_positions = Colour(field_area_positions, default=None)
        self.field_area = Colour(field_area, default=None)
        self.center_of_field_area = Colour(center_of_field_area, default=None)
        self.center_of_mass = Colour(center_of_mass, default=None)
        self.background = Colour(background, default=None)
        self.default_object_colour = Colour(default_object_colour,
                                           default=ImageColours.COL_DEFAULT_OBJECT)
        if opacity_guides is None:
            opacity_guides = ImageColours.OPACITY_GUIDES
        if opacity_guides < 0 or opacity_guides > 1:
            raise ValueError(f"opacity_guides ({opacity_guides}) has to be between 0 and 1")
        if opacity_object is None:
            opacity_object = ImageColours.OPACITY_OBJECT
        if opacity_object < 0 or opacity_object > 1:
            raise ValueError(f"opacity_object ({opacity_object}) has to be between 0 and 1")
        self.opacity_object = opacity_object
        self.opacity_guides = opacity_guides

    def as_dict(self):
        return OrderedDict(
            {"total_area": self.target_area.colour,
             "field_area_positions": self.field_area_positions.colour,
             "field_area": self.field_area.colour,
             "center_of_field_area": self.center_of_field_area.colour,
             "center_of_mass": self.center_of_mass.colour,
             "background": self.background.colour,
             "default_object": self.default_object_colour.colour,
             "object_opacity": self.opacity_object,
             "info_opacity": self.opacity_guides
             })

    def __str__(self):
        return misc.dict_to_text(self.as_dict(), col_a=24, col_b=10)



"""NOTE:
import always ethe ntire module and call `rng.generate` to ensure always the
access of the newly initialized random generator if  `init_random_generator` has
been called
"""

import numpy as np

generator = np.random.default_rng()


def init_random_generator(seed=None):
    """Init random generator and set random seed (optional)

    Parameters
    ----------
    seed: seed value
        must be  int, array_like[ints], SeedSequence, BitGenerator, Generator

    Notes
    -----
    see documentation of `numpy.random.default_rng()` of Python standard library
    """
    global generator
    generator = np.random.default_rng(seed=seed)
    if seed is not None:
        print("PyNSN seed: {}".format(seed))




__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import math
import numpy as np


def polar2cartesian(polar):
    """polar is an 2d-array representing polar coordinates (radius, angle)"""
    polar = np.asarray(polar)
    return np.array([polar[:, 0] * np.cos(polar[:, 1]),
                     polar[:, 0] * np.sin(polar[:, 1])]).T


def cartesian2polar(xy, radii_only=False):
    """polar coordinates (radius, angle)

    if only radii required you may consider radii_only=True for faster
    processing
    """
    xy = np.asarray(xy)
    radii = np.hypot(xy[:, 0], xy[:, 1])
    if radii_only:
        return radii
    else:
        return np.array([radii, np.arctan2(xy[:, 1], xy[:, 0])]).T


def cartesian2image_coordinates(xy, image_size):
    """convert cartesian to image coordinates with (0,0) at top left and
    reversed y axis
    """
    return (np.asarray(xy) * [1, -1]) + image_size / 2


def lines_intersect(line1, line2):
    # lines_overlap_on_x_axis
    x1, x2 = line1[0].x, line1[1].x
    x3, x4 = line2[0].x, line2[1].x
    e1_left, e1_right = min(x1, x2), max(x1, x2)
    e2_left, e2_right = min(x3, x4), max(x3, x4)
    lines_overlap_on_x_axis = (e1_left >= e2_left and e1_left <= e2_right) or \
            (e1_right >= e2_left and e1_right <= e2_right) or \
            (e2_left >= e1_left and e2_left <= e1_right) or \
            (e2_right >= e1_left and e2_right <= e1_right)

    # _lines_overlap_on_y_axis
    y1, y2 = line1[0].y, line1[1].y
    y3, y4 = line2[0].y, line2[1].y
    e1_top, e1_bot = min(y1, y2), max(y1, y2)
    e2_top, e2_bot = min(y3, y4), max(y3, y4)
    lines_overlap_on_y_axis = (e1_top >= e2_top and e1_top <= e2_bot) or \
           (e1_bot >= e2_top and e1_bot <= e2_bot) or \
           (e2_top >= e1_top and e2_top <= e1_bot) or \
           (e2_bot >= e1_top and e2_bot <= e1_bot)

    return lines_overlap_on_x_axis and lines_overlap_on_y_axis


# Gives distance if the point is facing edge, else False
def distance_between_edge_and_point(edge, point):
    # edge is a tuple of 2d coordinates
    if point_faces_edge(edge, point):
        area=triangle_area_at_points(edge[0], edge[1], point)
        base=edge[0].distance(edge[1])
        height=area/(0.5*base)
        return height
    return min(point.distance(edge[0]), point.distance(edge[1]))


def triangle_area_at_points(p1, p2, p3):
    # p1, p2, p3: 2d Coordinates
    a=p1.distance(p2)
    b=p2.distance(p3)
    c=p1.distance(p3)
    s=(a+b+c)/float(2)
    area=math.sqrt(s*(s-a)*(s-b)*(s-c))
    return area


# Finds angle using cos law
def angle(a, b, c):
    divid=float(a**2+b**2-c**2)
    divis=(2*a*b)
    if (divis)>0:
        result=divid/divis
        if result<=1.0 and result>=-1.0:
            return math.acos(result)
        return 0
    else:
        return 0


# Checks if point faces edge
def point_faces_edge(edge, point):
    a=edge[0].distance(edge[1])
    b=edge[0].distance(point)
    c=edge[1].distance(point)
    ang1, ang2 = angle(b, a, c), angle(c, a, b)
    if ang1>math.pi/2 or ang2>math.pi/2:
        return False
    return True


def center_of_positions(xy):
    min_max = np.array((np.min(xy, axis=0), np.max(xy, axis=0)))
    return np.reshape(min_max[1, :] - np.diff(min_max, axis=0) / 2, 2)



import numpy as _np
import numpy as np

from ._lib.misc import numpy_round2 as _numpy_round2
from ._lib import rng as _rng
from . import exceptions as _exceptions


def _round_samples(samples, round_to_decimals):
    if round_to_decimals is not None:
        return _numpy_round2(samples, decimals=round_to_decimals)
    else:
        return _np.array(samples)


class PyNSNDistribution(object):

    def __init__(self, min_max):
        if not isinstance(min_max, (list, tuple)) or len(min_max) != 2 or \
                (None not in min_max and min_max[0] > min_max[1]):
            raise TypeError("min_max {} ".format(min_max) + \
                            "has to be a tuple of two values (a, b) with a <= b.")
        self.min_max = min_max

    def as_dict(self):
        return {"distribution": type(self).__name__,
                "min_max": self.min_max}

    def sample(self, n, round_to_decimals=False):
        return NotImplementedError()

    def pyplot_samples(self, n=100000):
        try:
            from matplotlib.pyplot import hist, hist2d
        except:
            raise ImportError("To use pyplot_samples, please install matplotlib.")

        samples = self.sample(n=n)
        if samples.ndim == 2:
            return hist2d(samples[:,0], samples[:,1], bins=100)[2]
        else:
            return hist(samples, bins=100)[2]


class Uniform(PyNSNDistribution):
    """
    """

    def __init__(self, min_max):
        """Uniform distribution defined by the number range, min_max=(min, max)

        Parameter:
        ----------
        min_max : tuple (numeric, numeric)
            the range of the distribution
        """

        super().__init__(min_max)

    def sample(self, n, round_to_decimals=None):
        dist = _rng.generator.random(size=n)
        rtn = self.min_max[0] + dist * float(self.min_max[1] - self.min_max[0])
        return _round_samples(rtn, round_to_decimals)


class Levels(PyNSNDistribution):
    """
    """

    def __init__(self, levels, weights=None, exact_weighting=False):
        """Distribution of level. Samples from a population discrete categories
         with optional weights for each level or category.

        Parameter:
        ----------
        """
        super().__init__(min_max=(None, None))
        self.levels = levels
        self.exact_weighting = exact_weighting

        if weights is not None and len(levels) != len(weights):
            raise ValueError("Number weights does not match the number of levels")
        self.weights = weights

    def sample(self, n, round_to_decimals=None):
        if self.weights is None:
            p = _np.array([1 / len(self.levels)] * len(self.levels))
        else:
            p = _np.array(self.weights)
            p = p / _np.sum(p)

        if not self.exact_weighting:
            dist = _rng.generator.choice(a=self.levels, p=p, size=n)
        else:
            n_distr = n * p
            if not _np.alltrue(_np.round(n_distr) == n_distr):
                # problem: some n are floats
                try:
                    gcd = _np.gcd.reduce(self.weights)  # greatest common denominator
                    info = "\nSample size has to be a multiple of {}.".format(
                        int(_np.sum(self.weights / gcd)))
                except:
                    info = ""
                raise _exceptions.NoSolutionError(f"Can't find n={n} samples that" +
                      f" are exactly distributed as specified by the weights (p={p}). " +
                      info)

            dist = []
            for lev, n in zip(self.levels, n_distr):
                dist.extend([lev] * int(n))
            _rng.generator.shuffle(dist)

        return _round_samples(dist, round_to_decimals)

    def as_dict(self):
        d = super().as_dict()
        d.update({"population": self.levels,
                  "weights": self.weights,
                  "exact_weighting": self.exact_weighting})
        return d


class Triangle(PyNSNDistribution):
    """
    """

    def __init__(self, mode, min_max):
        super().__init__(min_max=min_max)
        self.mode = mode
        if (min_max[0] is not None and mode <= min_max[0]) or \
                (min_max[1] is not None and mode >= min_max[1]):
            txt = "mode ({}) has to be inside the defined min_max range ({})".format(
                mode, min_max)
            raise ValueError(txt)

    def sample(self, n, round_to_decimals=None):
        dist = _rng.generator.triangular(left=self.min_max[0], right=self.min_max[1],
                                         mode=self.mode, size=n)
        return _round_samples(dist, round_to_decimals)

    def as_dict(self):
        d = super().as_dict()
        d.update({"mode": self.mode})
        return d


class _PyNSNDistributionMuSigma(PyNSNDistribution):

    def __init__(self, mu, sigma, min_max):
        super().__init__(min_max)
        self.mu = mu
        self.sigma = abs(sigma)
        if (min_max[0] is not None and mu <= min_max[0]) or \
                (min_max[1] is not None and mu >= min_max[1]):
            txt = "mean ({}) has to be inside the defined min_max range ({})".format(
                mu, min_max)
            raise ValueError(txt)

    def as_dict(self):
        d = super().as_dict()
        d.update({"mu": self.mu,
                  "sigma": self.sigma})
        return d


class Normal(_PyNSNDistributionMuSigma):

    def __init__(self, mu, sigma, min_max=None):
        """Normal distribution with optional cut-off of minimum and maximum

        Resulting distribution has the defined mean and std, only if
        cutoffs values are symmetric.

        Parameter:
        ----------
        mu: numeric
        sigma: numeric
        min_max : tuple (numeric, numeric) or None
            the range of the distribution
        """
        if min_max is None:
            min_max = (None, None)
        super().__init__(mu, sigma, min_max)

    def sample(self, n, round_to_decimals=None):
        rtn = _np.array([])
        required = n
        while required > 0:
            draw = _rng.generator.normal(loc=self.mu, scale=self.sigma, size=required)
            if self.min_max[0] is not None:
                draw = _np.delete(draw, draw < self.min_max[0])
            if self.min_max[1] is not None:
                draw = _np.delete(draw, draw > self.min_max[1])
            if len(draw) > 0:
                rtn = _np.append(rtn, draw)
                required = n - len(rtn)
        return _round_samples(rtn, round_to_decimals)


class Normal2D(PyNSNDistribution):

    def __init__(self, mu, sigma, correlation, max_radius=None):
        """Two dimensional normal distribution with optional radial cut-off

        Resulting multivariate distribution has the defined means, standard
        deviations and correlation (approx.) between the two variables

        Parameter:
        ----------
        mu:  tuple of to numbers (numeric, numeric)
        sigma: tuple of to numbers  (numeric, numeric)
        correlation: numeric
            the correlation between a and b
        max_radius: numerical (optional)
            cut-off criterion: maximal distance from the distribution center (mu)
        """
        super().__init__(min_max=(None, None))

        try:
            lmu = len(mu)
            lsigma = len(sigma)
        except:
            lmu = None
            lsigma = None
        if lmu != 2 or lsigma != 2:
            raise ValueError("Mu and sigma has to be a tuple of two values.")
        if correlation < -1 or correlation > 1:
            raise ValueError("Correlations has to be between -1 and 1")

        self.mu = np.array(mu)
        self.sigma = np.abs(sigma)
        self.correlation = correlation
        self.max_radius = max_radius

    def varcov(self):
        """variance covariance matrix"""
        v = self.sigma ** 2
        cov = np.eye(2) * v
        cov[0, 1] = self.correlation * np.sqrt(v[0] * v[1])
        cov[1, 0] = cov[0, 1]
        return cov

    def sample(self, n, round_to_decimals=None):
        rtn = None
        required = n
        while required > 0:
            draw = _rng.generator.multivariate_normal(
                    mean=self.mu, cov=self.varcov(), size=required)
            if self.max_radius is not None:
                # remove to large radii
                r = np.hypot(draw[:,0], draw[:,1])
                draw = _np.delete(draw, r > self.max_radius, axis=0)

            if len(draw) > 0:
                # append
                if rtn is None:
                    rtn = draw
                else:
                    rtn = _np.append(rtn, draw, axis=0)
                required = n - len(rtn)

        return _round_samples(rtn, round_to_decimals)

    def as_dict(self):
        d = super().as_dict()
        d.update({"mu": self.mu.tolist(),
                  "sigma": self.sigma.tolist(),
                  "correlation": self.correlation,
                  "max_radius": self.max_radius})
        return d

    # TODO:Norma2D pyplot sample


class Beta(_PyNSNDistributionMuSigma):

    def __init__(self, mu=None, sigma=None, alpha=None, beta=None, min_max=None):
        """Beta distribution defined by the number range, min_max=(min, max),
         the mean and the standard deviation (std)

        Resulting distribution has the defined mean and std

        Parameter:
        ----------
        mu: numeric
        sigma: numeric
        min_max : tuple (numeric, numeric)
            the range of the distribution

        Note:
        -----
        Depending on the position of the mean in number range the
        distribution is left or right skewed.

        See Also:
        --------
        for calculated shape parameters [alpha, beta] see property
        `shape_parameter_beta`
        """
        if alpha is not None and beta is not None and (mu, sigma) == (None, None):
            mu, sigma = Beta._calc_mu_sigma(alpha, beta, min_max)
        elif mu is None or sigma is None or alpha is not None or beta is not None:
            raise TypeError("Either Mu & Sigma or Alpha & Beta have to specified.")
        super().__init__(mu=mu, sigma=sigma, min_max=min_max)

    def sample(self, n, round_to_decimals=None):
        if self.sigma is None or self.sigma == 0:
            return _np.array([self.mu] * n)

        alpha, beta = self.shape_parameter
        dist = _rng.generator.beta(a=alpha, b=beta, size=n)
        dist = (dist - _np.mean(dist)) / _np.std(dist)  # z values
        rtn = dist * self.sigma + self.mu
        return _round_samples(rtn, round_to_decimals)

    @property
    def shape_parameter(self):
        """Alpha (p) & beta (q) parameter for the beta distribution
        http://www.itl.nist.gov/div898/handbook/eda/section3/eda366h.htm

        Returns
        -------
        parameter: tuple
            shape parameter (alpha, beta) of the distribution

        """
        r = float(self.min_max[1] - self.min_max[0])
        m = (self.mu - self.min_max[0]) / r  # mean
        std = self.sigma / r
        x = (m * (1 - m) / std ** 2) - 1
        return x * m, (1 - m) * x

    @property
    def alpha(self):
        return self.shape_parameter[0]

    @property
    def beta(self):
        return self.shape_parameter[1]

    @staticmethod
    def _calc_mu_sigma(alpha, beta, min_max):
        a = float(alpha)
        b = float(beta)
        r = float((min_max[1] - min_max[0]))

        e = a / (a + b)
        mu = e * r + min_max[0]

        v = (a * b) / ((a + b) ** 2 * (a + b + 1))
        sigma = _np.sqrt(v) * r
        return mu, sigma



"""
Dot Array Sequence
"""
__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

#FIXME Depricated model currently BROKEN
# FIXME BROKEN MODULE

from hashlib import md5 as _md5
import numpy as _np

from .._lib import misc as _misc
from .._lib import DotArray
from ..visual_properties import flags, fit


class DASequence(object):

    def __init__(self):
        """ docu the use of numerosity_idx see get_array_numerosity
        dot array, please use append_dot_array and delete_array to modify the list

        """

        self.dot_arrays = []
        self.method = None
        self.error = None
        self.numerosity_idx = {}

    def append_dot_arrays(self, arr):
        if isinstance(arr, DotArray):
            arr = [arr]
        self.dotextend(arr)
        self.numerosity_idx = {da._properties.numerosity: idx for idx, da in enumerate(self.dot_arrays)}

    def delete_dot_arrays(self, array_id):
        self.dotpop(array_id)
        self.numerosity_idx = {da._properties.numerosity: idx for idx, da in enumerate(self.dot_arrays)}

    def get_array(self, numerosity):
        """returns array with a particular numerosity"""

        try:
            return self.dot_arrays[self.numerosity_idx[numerosity]]
        except:
            return None

    @property
    def min_max_numerosity(self):
        return self.dot_arrays[0].properties.numerosity, \
               self.dot_arrays[-1].properties.numerosity

    @property
    def hash(self):
        """meta hash of object ids"""

        m = _md5()
        for da in self.dot_arrays:
            m.update(da.hash.encode("UTF-8"))
        return m.hexdigest()

    def get_properties_dict(self):
        """dictionary with arrays

        Examples
        --------
        making a pandas dataframe with aa properties
        >>> d = my_da_sequence.as_dict()
        >>> array = []
        >>> for x in range(len(d["Hash"])):
        >>>    row = map(lambda k: d[k][x], d.keys())
        >>>    array.append(list(row))
        >>> return pandas.DataFrame(array, columns=list(d.keys()))
        """
        dicts = [x._properties.as_dict() for x in self.dot_arrays]
        rtn = _misc.join_dict_list(dicts)
        rtn['sequence_id'] = [self.hash] * len(self.dot_arrays)  # all arrays have the same _sequence ID
        return rtn


    def get_numerosity_correlations(self):
        feat = self.get_properties_dict()
        del feat['hash']
        prop_np = _np.round(_np.asarray(feat.values()).T, 2)
        cor = _np.corrcoef(prop_np, rowvar=False)
        cor = cor[0, :]
        names = feat.keys()
        rtn = {}
        for x in range(1, len(cor)):
            rtn[names[x]] = cor[x]
        return rtn

    def __str__(self):
        return self.get_csv()

    def get_csv(self, variable_names=True, colour_column=False,
                hash_column=True):

        rtn = ""
        tmp_var_names = variable_names

        for da in self.dot_arrays:
            rtn += da.csv(num_idx_column=True, hash_column=False,
                          variable_names=tmp_var_names,
                          colour_column=colour_column)
            tmp_var_names = False

        if hash_column:
            obj_id = self.hash
            rtn2 = ""
            tmp_var_names = variable_names
            for l in rtn.split("\n"):
                if tmp_var_names:
                    rtn2 += "hash," + l + "\n"
                    tmp_var_names = False
                elif len(l) > 0:
                    rtn2 += "{},{}\n".format(obj_id, l)
            return rtn2
        else:
            return rtn


def create(specs,
           adapt_property,
           adapt_value,
           min_max_numerosity,
           round_decimals = None,
           source_number = None):  # todo could be an iterator
    """factory function

    Methods takes take , you might use make Process
        adapt_properties:
                continuous property or list of continuous properties to be adapt
                or None
     returns False is error occurred (see self.error)
    """
    try:
        l = len(min_max_numerosity)
    except:
        l = 0
    if l != 2:
        raise ValueError("min_max_numerosity has to be a pair of (min, max)")

    min_, max_ = sorted(min_max_numerosity)

    if source_number is None:
        if adapt_property in [flags.SPARSITY]:
            source_number = min_
        elif adapt_property in [flags.FIELD_AREA, flags.COVERAGE] :
            source_number = max_
        else:
            source_number = min_ + ((max_ - min_)//2)

    check_property_list(adapt_property)

    if not isinstance(specs, DotArraySpecs):
        raise TypeError("Specs has to be of type DotArraySpecs, but not {}".format(
            type(specs).__name__))

    # keep field area
    if adapt_property in list(flags.SPACE_FEATURES) + [flags.COVERAGE]:
        prefer_keeping_field_area = True
    else:
        prefer_keeping_field_area = False

    # make source image
    if source_number is None:
        source_number = min_ + int((max_ - min_)/2)
    source_da = random_array.create(n_objects=source_number,
                                    specs=specs)
    source_da = fit.visual_property(source_da, property_flag=adapt_property, value=adapt_value)
    source_da.center_array()
    source_da.mod_round_values(round_decimals)

    # adapted deviants
    rtn = DASequence()
    rtn.method = adapt_property

    # decreasing
    if min_ < source_number:
        tmp, error = _make_adapted_deviants(
            reference_da=source_da,
            adapt_property=adapt_property,
            target_numerosity=min_,
            round_decimals=round_decimals,
            prefer_preserve_field_area=prefer_keeping_field_area)

        rtn.append_dot_arrays(list(reversed(tmp)))
        if error is not None:
            rtn.error = error
    # source number
    rtn.append_dot_arrays(source_da)
    # increasing
    if max_ > source_number:
        tmp, error = _make_adapted_deviants(
            reference_da=source_da,
            adapt_property=adapt_property,
            target_numerosity=max_,
            round_decimals=round_decimals,
            prefer_preserve_field_area=prefer_keeping_field_area)
        rtn.append_dot_arrays(tmp)
        if error is not None:
            rtn.error = error

    return rtn

def _make_adapted_deviants(reference_da, adapt_property, target_numerosity,
                           round_decimals, prefer_preserve_field_area):
    """helper function. Do not use this method. Please use make"""



    if reference_da._properties.numerosity == target_numerosity:
        change = 0
    elif reference_da._properties.numerosity > target_numerosity:
        change = -1
    else:
        change = 1

    da = reference_da.copy()
    da_sequence = []

    error = None
    #print(adapt_props, target_numerosity)
    while True:
        try:
            da = da.get_number_deviant(change_numerosity=change,
                                       preserve_field_area=prefer_preserve_field_area)
        except:
            return [], "ERROR: Can't find the a make adapted deviants"

        da = fit.visual_property(da, property_flag=adapt_property,
                                 value=reference_da.properties.get(adapt_property))
        cnt = 0
        while True:
            cnt += 1
            ok, mesg = da.realign()
            if ok:
                break
            if cnt > 10:
                error = u"ERROR: realign, " + str(cnt) + ", " + str(da._properties.numerosity)

        da.mod_round_values(round_decimals)
        da_sequence.append(da)

        if error is not None or da._properties.numerosity == target_numerosity:
            break

    return da_sequence, error


def check_property_list(feature_list):
    """helper function
    raises TypeError or Runtime errors if checks fail
    * type check
    * dependency check
    """

    size_occured = ""
    space_occured = ""
    error = "Incompatible properties to adapt: {} & {}"

    if not isinstance(feature_list, (tuple, list)):
        feature_list = [feature_list]

    for x in feature_list:
        if x not in flags:
            raise ValueError("Parameter is not a continuous propertyor a " + \
                            "list of continuous properties")
            # continious property or visual prop

        if x in flags.SIZE_FEATURES:
            if len(size_occured)>0:
                raise ValueError(error.format(x, size_occured))
            else:
                size_occured = x

        if x in flags.SPACE_FEATURES:
            if len(space_occured)>0:
                raise RuntimeError(error.format(x, space_occured))
            else:
                space_occured = x

